"""
w2lapp.addmodifyform.py: input form for adding and modifying an entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: addmodifyform.py,v 1.207 2009/08/07 09:29:08 michael Exp $
"""

import re,sets,msbase,pyweblib,ldap,ldif,ldap.schema,\
       ldaputil.schema,ldapsession,w2lapp.core,w2lapp.cnf,w2lapp.form,w2lapp.gui,w2lapp.read,w2lapp.schema
from ldaputil.base import ParentDN,rdn_dict
from msbase import listremove

from w2lapp.core import utf2display
from w2lapp.schema.viewer import displayNameOrOIDList

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO


heading_msg = {'modifyform':'Modify entry','addform':'Add new entry'}

attrtype_pattern = r'[\w;.]+(;[\w_-]+)*'
attrvalue_pattern = r'(([^,]|\\,)*|".*?")'
rdn_pattern = attrtype_pattern + r'[ ]*=[ ]*' + attrvalue_pattern
dn_pattern   = rdn_pattern + r'([ ]*,[ ]*' + rdn_pattern + r')*[ ]*'
ldif.dn_regex   = re.compile('^%s$' % dn_pattern)


class UserEditableEntry(ldaputil.schema.Entry):
  """
  Base class for a schema-aware entry containing
  solely attributes editable by a user
  """

  def __init__(self,ls,sub_schema,dn,ldap_entry):
    self._ls = ls
    ldaputil.schema.Entry.__init__(self,sub_schema,dn,ldap_entry)

  def get_object_class_oids(self):
    return dict([
      (self._s.name2oid[ldap.schema.models.ObjectClass].get(o,o),None)
      for o in self.get('objectClass',[])
    ])

  def attribute_types(self):
    # Initialize a list of assertions for filtering attribute types
    # displayed in the input form
    attr_type_filter = [
      ('no_user_mod',[0]),
#        ('usage',range(2)),
      ('collective',[0]),
    ]
    # Check whether Manage DIT control is in effect,
    # let python-ldap filter out OBSOLETE attribute types otherwise
    relax_rules_enabled = self._ls.l._get_server_ctrls('__write__').has_key(ldapsession.CONTROL_RELAXRULES)
    if not relax_rules_enabled:
      attr_type_filter.append(('obsolete',[0]))

    # Filter out extensibleObject
    object_class_oids = self.get_object_class_oids()
    try:
      del object_class_oids['1.3.6.1.4.1.1466.101.120.111']
    except KeyError:
      try:
        del object_class_oids['extensibleObject']
      except KeyError:
        pass

    required_attrs_dict,allowed_attrs_dict = self._s.attribute_types(
      object_class_oids.keys(),
      attr_type_filter=attr_type_filter,
      raise_keyerror=0,
      ignore_dit_content_rule=relax_rules_enabled
    )
      
    # Additional check whether to explicitly add object class attribute.
    # This is a work-around for LDAP servers which mark the
    # objectClass attribute as not modifiable (e.g. MS Active Directory)
    if not required_attrs_dict.has_key('2.5.4.0') and \
       not allowed_attrs_dict.has_key('2.5.4.0'):
      required_attrs_dict['2.5.4.0'] = self._s.get_obj(ldap.schema.ObjectClass,'2.5.4.0')
    return required_attrs_dict,allowed_attrs_dict


class InputFormEntry(ldaputil.schema.Entry):

  def __init__(self,sid,form,ls,dn,schema,entry):
    assert type(dn)==type(u'')
    ldaputil.schema.Entry.__init__(self,schema,None,entry)
    self.sid = sid
    self.form = form
    self.ls = ls
    self.entry = entry
    self.set_dn(dn)
    self.relax_rules_enabled = ls.l._get_server_ctrls('__write__').has_key(ldapsession.CONTROL_RELAXRULES)

  def set_dn(self,dn):
    self.dn = dn
    self.rdn_dict = self._get_rdn_dict(dn)

  def _get_rdn_dict(self,dn):
    assert type(dn)==type(u'')
    entry_rdn_dict = ldap.schema.Entry(self._s,None,rdn_dict(dn,self.ls.charset))
    for attr_type,attr_values in entry_rdn_dict.items():
      del entry_rdn_dict[attr_type]
      d = ldap.cidict.cidict()
      for attr_value in attr_values:
        assert type(attr_value)==type('')
        d[attr_value] = None
      entry_rdn_dict[attr_type] = d
    return entry_rdn_dict

  def __getitem__(self,nameoroid):
    """
    Return HTML input field(s) for the attribute specified by nameoroid.
    """

    oid = self._at2key(nameoroid)[0]
    try:
      attr_values = ldap.schema.Entry.__getitem__(self,nameoroid)
    except KeyError:
      attr_values = []

    # Attribute value list must contain one element
    # to display at least display one input field
    attr_values = attr_values or [None]

    result = []
    nameoroid_se = self._s.get_obj(ldap.schema.AttributeType,oid)
    syntax_class = w2lapp.schema.syntaxes.syntax_registry.syntaxClass(self._s,nameoroid)

    attr_type_field_name,input_field_name = {
      1:('in_attrtype','in_value'),
      0:('in_binattrtype','in_binvalue'),
    }[syntax_class.editable]

    # Eliminate binary attribute values from input form
    if not syntax_class.editable:
      attr_values = ['']

    for attr_value in attr_values:

      attr_inst = syntax_class(self.sid,self.form,self.ls,self.dn,self._s,nameoroid,attr_value,self.entry)

      if oid=='2.5.4.0' or \
         ( attr_value and \
           self.rdn_dict.has_key(nameoroid) and \
           self.rdn_dict[nameoroid].has_key(attr_value)
         ) or \
        (not self.relax_rules_enabled and w2lapp.schema.no_userapp_attr(self._s,oid)):
        result.append('\n'.join([
          w2lapp.gui.HIDDEN_FIELD % (attr_type_field_name,nameoroid.encode('ascii'),''),
          w2lapp.gui.HIDDEN_FIELD % (
            input_field_name,
            utf2display(self.form.accept_charset,attr_inst.formValue(),sp_entity='  '),
            utf2display(self.form.accept_charset,attr_inst.formValue(),sp_entity='&nbsp;&nbsp;')
          )
        ]))

      else:
        attr_title = u''
        attr_type_tags = []
        if nameoroid_se:
          attr_title = unicode(nameoroid_se.desc or '','utf-8')
          # Determine whether transfer syntax has to be specified with ;binary
          if w2lapp.schema.NEEDS_BINARY_TAG.has_key(oid) or \
             w2lapp.schema.NEEDS_BINARY_TAG.has_key(nameoroid_se.syntax):
            attr_type_tags = [';binary']
        input_field = attr_inst.formField()
        input_field.name = input_field_name
        input_field.charset = self.form.accept_charset
        result.append('\n'.join([
          w2lapp.gui.HIDDEN_FIELD % (
            attr_type_field_name,
            ''.join((nameoroid,';'.join(attr_type_tags))),
            ''
          ),
          input_field.inputHTML(title=attr_title)
        ]))

    return '<br>\n'.join(result)


def PrintInputFieldset(
  outf,sid,form,dn,sub_schema,
  attr_types_dict,ignore_attr_types_dict,
  fieldset_title,entry,hidden=0
):
  outf_lines = []
  outf_lines.append("""<fieldset title="%s">
    <legend>%s</legend>
    <table summary="%s">
    """ % (fieldset_title,fieldset_title,fieldset_title)
  )
  seen_attr_type_oids = ldap.cidict.cidict()
  attr_type_names = ldap.cidict.cidict()
  for a in entry.keys():
    at_oid = entry._at2key(a)[0]
    if not ignore_attr_types_dict.has_key(at_oid) and attr_types_dict.has_key(at_oid):
      seen_attr_type_oids[at_oid] = None
      attr_type_names[a.encode('ascii')] = None
  for at_oid,at_se in attr_types_dict.items():
    if at_se and \
       not seen_attr_type_oids.has_key(at_oid) and \
       not ignore_attr_types_dict.has_key(at_oid) and \
       not w2lapp.schema.no_userapp_attr(sub_schema,at_oid):
        attr_type_names[(at_se.names or (at_se.oid,))[0].encode('ascii')] = None
  attr_types = attr_type_names.keys()
  attr_types.sort(msbase.CaseinsensitiveStringCompare)
  for attr_type in attr_types:
    attr_type_name = w2lapp.gui.SchemaElementName(sid,form,dn,sub_schema,attr_type,ldap.schema.AttributeType)
    attr_value_field_html = entry[attr_type]
    outf_lines.append('<tr>\n<td>\n%s\n</td>\n<td>\n%s\n</td>\n</tr>\n' % (attr_type_name,attr_value_field_html))
  outf_lines.append('</table></fieldset>')
  outf.write('\n'.join(outf_lines))

  return # PrintInputFieldset()


def ObjectClassForm(
  sid,outf,form,command,ls,sub_schema,dn,rdn,
  existing_object_classes,structural_object_class
):
  """Form for choosing object class(es)"""

  ObjectClass = ldap.schema.ObjectClass
  all_oc = [
    (sub_schema.get_obj(ObjectClass,oid).names or (oid,))[0]
    for oid in sub_schema.listall(ObjectClass)
  ]

  relax_rules_enabled = ls.l._get_server_ctrls('__write__').has_key(ldapsession.CONTROL_RELAXRULES)

  command_hidden_fields = [('dn',dn)]

  if rdn:
    command_hidden_fields.append(('add_rdn',rdn))
  existing_structural_oc,existing_abstract_oc,existing_auxiliary_oc = w2lapp.schema.object_class_categories(sub_schema,existing_object_classes)
  all_structural_oc,all_abstract_oc,all_auxiliary_oc = w2lapp.schema.object_class_categories(sub_schema,all_oc)

  dit_structure_rule_html=''

  if sub_schema.sed[ldap.schema.models.DITStructureRule]:

    # Determine possible structural object classes based on DIT structure rules
    # and name forms if DIT structure rules are defined in subschema
    if command=='addform':
      parent_dn = dn
    else:
      parent_dn = ldaputil.base.ParentDN(dn)
    dit_structure_ruleid = ls.getGoverningStructureRule(parent_dn,sub_schema)
    if dit_structure_ruleid!=None:
      subord_structural_ruleids,subord_structural_oc = sub_schema.get_subord_structural_oc_names(dit_structure_ruleid)
      if subord_structural_oc:
        all_structural_oc = subord_structural_oc
        dit_structure_rule_html = 'DIT structure rules:<br>%s' % ('<br>'.join(
            displayNameOrOIDList(sid,form,dn,sub_schema,subord_structural_ruleids,ldap.schema.models.DITStructureRule)
        ))

  existing_misc_oc = sets.Set(existing_object_classes)
  for a in existing_structural_oc+existing_abstract_oc+existing_auxiliary_oc:
    existing_misc_oc.discard(a)
  existing_misc_oc = list(existing_misc_oc)

  dit_content_rule_html = ''
  # Try to look up a DIT content rule
  if existing_object_classes and structural_object_class:
    # Determine OID of structural object class
    soc_oid = sub_schema.name2oid[ldap.schema.ObjectClass].get(structural_object_class,structural_object_class)
    dit_content_rule = sub_schema.get_obj(ldap.schema.DITContentRule,soc_oid,None)
    if dit_content_rule!=None:
      if dit_content_rule.obsolete:
        dit_content_rule_status_text = 'Ignored obsolete'
      elif relax_rules_enabled:
        dit_content_rule_status_text = 'Ignored'
      else:
        dit_content_rule_status_text = 'Governed by'
        if dit_content_rule.aux:
          all_auxiliary_oc = list(dit_content_rule.aux)
      dit_content_rule_html = '%s<br>DIT content rule:<br>%s' % (
        dit_content_rule_status_text,
        w2lapp.gui.SchemaElementName(sid,form,dn,sub_schema,dit_content_rule.names[0],ldap.schema.DITContentRule)
      )

  abstract_select_field = w2lapp.form.ObjectClassSelect(
    name='ldap_oc',text='Abstract object class(es)',options=all_abstract_oc,default=existing_abstract_oc,size=20
  )
  structural_select_field = w2lapp.form.ObjectClassSelect(
    name='ldap_oc',text='Structural object class(es)',options=all_structural_oc,default=existing_structural_oc,size=20
  )
  auxiliary_select_field = w2lapp.form.ObjectClassSelect(
    name='ldap_oc',text='Auxiliary object class(es)',options=all_auxiliary_oc,default=existing_auxiliary_oc,size=20
  )
  misc_select_field = w2lapp.form.ObjectClassSelect(
    name='ldap_oc',text='Misc. object class(es)',options=[],default=existing_misc_oc,size=20
  )
  if existing_misc_oc:
    misc_select_field_th = '<th><label for="add_misc_oc">Misc.<label></th>'
    misc_select_field_td = '<td>%s</td>' % (misc_select_field.inputHTML(id_value='add_misc_oc'))
  else:
    misc_select_field_th = ''
    misc_select_field_td = ''

  # Build an select field based on config param 'addform_entry_templates'
  if command=='addform':
    addform_entry_templates_keys = w2lapp.cnf.GetParam(ls,'addform_entry_templates',{}).keys()
    addform_entry_templates_keys.append('')
    addform_entry_templates_keys.sort()
    add_template_select_field = pyweblib.forms.Select(
      'add_template',
      u'LDIF template',
      1,
      options=addform_entry_templates_keys,
      default=''
    )
    add_template_field_html = '<label for="add_template_name">%s:<label>%s' % (
      utf2display(form.accept_charset,add_template_select_field.text),
      add_template_select_field.inputHTML(
        default='',
        id_value='add_template_name',
        title='LDIF templates for new entries including object classes'
    ))

  else:
    add_template_field_html = ''

  Msg = {
    'addform':'Choose object class(es) for new entry or choose from quick-list.',
    'modifyform':'You may change the object class(es) for the entry.',
  }[command]

  context_menu_list = w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  context_menu_list.append(form.applAnchor('oid','Browse Schema',sid,[('dn',dn)]))

  # Write HTML output of object class input form
  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,
    'Object Class Selection',
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=context_menu_list
  )

  outf.write("""
    <div id="Input" class="Main">\n<h1>%s</h1>
      %s
       %s
       <p>
         %s
         <label for="add_form_type">Form type:</label> %s
         <input type="submit" value="Next &gt;&gt;">
       </p>
       <p>%s</p>
       <table summary="Choosing different categories of object classes for the entry">
         <tr>
           <th><label for="add_structural_oc">Structural</label></th>
           <th><label for="add_auxiliary_oc">Auxiliary</label></th>
           <th><label for="add_abstract_oc">Abstract</label></th>
           %s
         </tr>
         <tr>
           <td><label for="add_structural_oc">%s</label></td>
           <td><label for="add_auxiliary_oc">%s</label></td>
           <td><label for="add_abstract_oc">%s</label></td>
           %s
         </tr>
         <tr>
           <td>%s</td>
           <td>%s</td>
         </tr>
       </table>
      </form>
    </div>
  """ % (
    heading_msg[command],
    form.beginFormHTML(command,sid,'GET',None),
    ''.join([
      w2lapp.gui.HIDDEN_FIELD % (param_name,utf2display(form.accept_charset,param_value,sp_entity='  '),'')
      for param_name,param_value in command_hidden_fields
    ]),
    add_template_field_html,
    form.field['input_formtype'].inputHTML(id_value='add_form_type'),
    Msg,
    misc_select_field_th,
    structural_select_field.inputHTML(id_value='add_structural_oc',title='Structural object classes to be added'),
    auxiliary_select_field.inputHTML(id_value='add_auxiliary_oc',title='Auxiliary object classes to be added'),
    abstract_select_field.inputHTML(id_value='add_abstract_oc',title='Abstract object classes to be added'),
    misc_select_field_td,
    dit_structure_rule_html,
    dit_content_rule_html,
  ))
  w2lapp.gui.PrintFooter(outf,form)
  return # ObjectClassForm()


def ReadLDIFTemplate(ls,template_name):
  addform_entry_templates = w2lapp.cnf.GetParam(ls,'addform_entry_templates',{})
  if not addform_entry_templates.has_key(template_name):
    raise w2lapp.core.ErrorExit('LDIF template key not known.')
  ldif_file_name = addform_entry_templates[template_name]
  ldif_file = open(ldif_file_name,'r')
  try:
    ldif_parser = ldif.LDIFRecordList(
      ldif_file,
      ignored_attr_types=[],
      max_entries=1,
      process_url_schemes=w2lapp.cnf.misc.ldif_url_schemes
    )
    ldif_parser.parse()
  except IOError:
    raise w2lapp.core.ErrorExit('Error reading LDIF template.')
  dn,entry = ldif_parser.all_records[0]
  return dn,entry # ReadLDIFTemplate()


def ConfiguredConstantAttributes(ls):
  """
  Build a list of attributes assumed in configuration to be constant while editing entry
  """
  modify_constant_attrs = ldap.cidict.cidict()
  modify_constant_attrs.update(
    {}.fromkeys(
      w2lapp.cnf.GetParam(
        ls,'modify_constant_attrs',
        ['createTimestamp','modifyTimestamp','creatorsName','modifiersName']
      )
    )
  )
  return modify_constant_attrs # ConfiguredConstantAttributes()


def AssertionFilter(ls,entry):
  assertion_filter_list = []
  for attr_type in ConfiguredConstantAttributes(ls).keys():
    try:
      attr_values = entry[attr_type]
    except KeyError:
      continue
    else:
      assertion_filter_list.extend([
        '(%s)' % ('='.join((attr_type,ldap.filter.escape_filter_chars(attr_value,escape_mode=1))))
        for attr_value in attr_values
      ])
  if assertion_filter_list:
    return '(&%s)' % ''.join(assertion_filter_list)
  else:
    return '(objectClass=*)'


def ReadOldEntry(ls,dn,sub_schema,assertion_filter=None):
  """
  Retrieve all editable attribute types an entry
  """
  assertion_filter = assertion_filter or '(objectClass=*)'
  # Build a list of attributes to be read
  read_attrs = ldap.cidict.cidict({'*':None})
  read_attrs.update(ConfiguredConstantAttributes(ls))
  read_attrs.update({}.fromkeys(w2lapp.cnf.GetParam(ls,'requested_attrs',[])))
  # Read the editable attribute values of entry
  try:
    ldap_entry = ls.readEntry(dn,read_attrs.keys(),search_filter=assertion_filter.encode(ls.charset),no_cache=1)[0][1]
  except IndexError:
    raise ldap.NO_SUCH_OBJECT
  entry = UserEditableEntry(ls,sub_schema,dn.encode(ls.charset),ldap_entry)
  return entry # ReadOldEntry()


def MultiValueAttributeTypes(sub_schema,input_form_entry,required_attrs_dict,allowed_attrs_dict):
  # Build two attribute type lists for normal input and file input
  all_txtattrtypes = []
  all_binattrtypes = []
  all_attrs_dict = {}
  object_class_oids = ldap.cidict.cidict(input_form_entry.get_object_class_oids())
  # Special treatment for object class extensibleObject
  if object_class_oids.has_key('1.3.6.1.4.1.1466.101.120.111') or \
     object_class_oids.has_key('extensibleObject'):
    all_attrs_dict.update(sub_schema.sed[ldap.schema.models.AttributeType])
  else:      
    all_attrs_dict.update(required_attrs_dict)
    all_attrs_dict.update(allowed_attrs_dict)
  # Sort out all non-editable attribute types including 'objectClass'
  for oid,a in all_attrs_dict.items():
    if not a or a.oid=='2.5.4.0' or a.single_value or a.collective or \
       w2lapp.schema.no_userapp_attr(sub_schema,a.oid):
      del all_attrs_dict[oid]
  # Sort into textual and binary attributes
  for a in all_attrs_dict.values():
    a2 = (a.names or (a.oid,))[0]
    if w2lapp.schema.no_humanreadable_attr(sub_schema,a.oid):
      all_binattrtypes.append(a2)
    else:
      all_txtattrtypes.append(a2)
  all_attrtypes = all_attrs_dict.keys()
  all_attrtypes.sort(msbase.CaseinsensitiveStringCompare)
  all_txtattrtypes.sort()
  all_binattrtypes.sort(msbase.CaseinsensitiveStringCompare)
  return all_txtattrtypes,all_binattrtypes # EditableAttributeTypeLists()


def WriteInputFormFooter(outf,form,sub_schema,input_form_entry,required_attrs_dict,allowed_attrs_dict,input_formtype):

  all_txtattrtypes,all_binattrtypes = MultiValueAttributeTypes(sub_schema,input_form_entry,required_attrs_dict,allowed_attrs_dict)

  all_attrtypes = ['']
  all_attrtypes.extend(all_txtattrtypes)
  all_attrtypes.extend(all_binattrtypes)
  all_attrtypes.sort(msbase.CaseinsensitiveStringCompare)
  
  if all_txtattrtypes:
    all_txtattrtypes_html = """        <tr><td title="Choose additional textual attribute type">
          <select name="input_addattrtype"><option>%s</option></select>
          <input name="input_formtype" value="+" type="submit">
        </td></tr>
    """ % ('</option>\n<option>'.join(all_txtattrtypes))
  else:
    all_txtattrtypes_html = ''

  if all_binattrtypes:
    all_binattrtypes_html = """
        <tr><td title="Additional binary attribute type and value">
          <select name="in_binattrtype"><option>%s</option></select>
          <input type="file" name="in_binvalue" size="40" maxlength="%d" value="">
        </td></tr>
    """ % ('</option>\n<option>'.join(all_binattrtypes),w2lapp.cnf.misc.input_maxfilelen)
  else:
    all_binattrtypes_html = ''

  outf.write("""
    <fieldset>
      <legend>Add more values to multi-valued attributes</legend>
      <table summary="More values to multi-valued attributes">
        %s
        %s
        <tr><td title="Additional textual attribute type and value">
          <input name="in_attrtype" size="15" maxlength="200" value="">
          <input name="in_value" size="40" value="">
        </td></tr>
        <tr><td title="Additional binary attribute type and value">
          <input name="in_binattrtype" size="15" maxlength="200" value="">
          <input type="file" name="in_binvalue" size="40" maxlength="%d" value="">
        </td></tr>
      </table>
      <input type="hidden" name="input_currentformtype" value="%s">
    </fieldset>
""" % (
    all_txtattrtypes_html,
    all_binattrtypes_html,
    w2lapp.cnf.misc.input_maxfilelen,
    input_formtype,
  ))


def WriteLDIFField(outf,form,ls,sub_schema,dn,entry,required_attrs_dict,allowed_attrs_dict):
  f = StringIO()
  ldif_writer = ldif.LDIFWriter(f)
  ldap_entry = {}
  for attr_type,attr_values in entry.items():
    ldap_entry[attr_type] = [
      attr_value
      for attr_value in attr_values
    ]
  ldif_writer.unparse(dn.encode(ls.charset),ldap_entry)
  outf.write("""
    <fieldset>
      <legend>Raw LDIF data</legend>
      <textarea name="in_ldif" rows="50" cols="80">\n%s</textarea>
      <p>
        Notes:
      </p>
      <ul>
        <li>Lines containing "dn:" will be ignored</li>
        <li>Only the first entry (until first empty line) will be accepted</li>
        <li>Maximum length is set to %d bytes</li>
        <li>Allowed URL schemes: %s</li>
      </ul>
    </fieldset>
""" % (
    utf2display(form.accept_charset,f.getvalue()),
    w2lapp.cnf.misc.ldif_maxbytes,
    ', '.join(w2lapp.cnf.misc.ldif_url_schemes))
  )


def w2l_AddForm(sid,outf,command,form,ls,dn,Msg='',rdn_default='',entry=None,skip_oc_input=0):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  if 'ldap_oc' in form.inputFieldNames:
    # Read objectclass(es) from input form
    entry = {'objectClass':[
      oc.encode('ascii')
      for oc in form.field['ldap_oc'].value
    ]}
    add_rdn = None

  elif 'add_template' in form.inputFieldNames:
    add_rdn,entry = ReadLDIFTemplate(ls,form.field['add_template'].value[0])

  elif not skip_oc_input:
    # Output the web page with object class input form
    ObjectClassForm(
      sid,outf,form,'addform',ls,sub_schema,dn,
      form.getInputValue('add_rdn',[rdn_default])[0],
      [],None
    )
    return

  else:
    add_rdn = form.getInputValue('add_rdn',[rdn_default])[0]
    

  entry = UserEditableEntry(ls,sub_schema,dn.encode(ls.charset),entry or {})

  rdn_options = entry.get_rdn_templates()

  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,
    heading_msg[command],
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

  input_formtype = form.getInputValue('input_formtype',['Template'])[0]
  if not input_formtype in ['Template','Table','LDIF']:
    input_formtype = form.getInputValue('input_currentformtype',['Template'])[0]

  # Check whether to fall back to table input mode
  if input_formtype=='Template':
    template_oc,read_template_dict = w2lapp.read.DetermineHTMLTemplates(ls,entry,'input_template')
    if not template_oc:
      Msg = '<br>'.join((Msg,'<strong>Info: No templates defined for chosen object classes.</strong>'))
      input_formtype='Table'

  if entry.get_structural_oc() is None:
    Msg = '<br>'.join((Msg,'<strong>Warning: You did not choose a STRUCTURAL object class!</strong>'))

  outf.write("""<div id="Input" class="Main">
    <h1>%s</h1>
    %s
    %s
    <input type="hidden" name="dn" value="%s">
    <input type="submit" value="Add">
    Change input form:
    <input type="submit" name="input_formtype" value="%s">
    <input type="submit" name="input_formtype" value="%s">
    <br>
    """ % (
      heading_msg[command],
      Msg,
      form.beginFormHTML('add',sid,'POST',target=None,enctype='multipart/form-data'),
      dn.encode(form.accept_charset),
      {'Template':'Table','Table':'Template','LDIF':'Template'}[input_formtype],
      {'Template':'LDIF','Table':'LDIF','LDIF':'Table'}[input_formtype],
    )
  )

  required_attrs_dict,allowed_attrs_dict = entry.attribute_types()

  if rdn_options and len(rdn_options)>0:
    # <select> field
    f = pyweblib.forms.Select('add_rdn','RDN variants',1,options=rdn_options)   
  else:
    # Just a normal <input> text field
    f = form.field['add_rdn']
  if add_rdn:
    f.setDefault(add_rdn)
  elif len(required_attrs_dict)<=2:
    for at_oid in required_attrs_dict.keys():
      if at_oid!='2.5.4.0':
        rdn_at_name = (required_attrs_dict[at_oid].names or (at_oid,))[0]
        f.setDefault(rdn_at_name+'=')
  outf.write('<p>RDN: %s</p>' % f.inputHTML())

  input_form_entry = InputFormEntry(sid,form,ls,dn,sub_schema,entry)

  # Use HTML templates for displaying a partial input form
  if input_formtype=='Template':
    displayed_attrs = w2lapp.read.PrintTemplateOutput(sid,outf,form,ls,sub_schema,entry,input_form_entry,'input_template')
    outf_lines = []
    for attr_type,attr_values in entry.items():
      at_oid = entry._at2key(attr_type)[0]
      if not at_oid in displayed_attrs:
        syntax_class = w2lapp.schema.syntaxes.syntax_registry.syntaxClass(sub_schema,attr_type)
        for attr_value in attr_values:
          attr_inst = syntax_class(sid,form,ls,dn,sub_schema,attr_type,attr_value,entry)
          outf_lines.append(w2lapp.gui.HIDDEN_FIELD % (
            'in_attrtype',utf2display(form.accept_charset,attr_type),''
          ))
          try:
            attr_value_html = utf2display(form.accept_charset,attr_inst.formValue())
          except UnicodeDecodeError:
            # Simply display an empty string if anything goes wrong with Unicode decoding (e.g. with binary attributes)
            attr_value_html = ''
          outf_lines.append(w2lapp.gui.HIDDEN_FIELD % (
            'in_value',attr_value_html,''
          ))
    outf.write(''.join(outf_lines))

  elif input_formtype=='Table':
    # Displaying rest of input form as simple list
    displayed_attrs = {}
    PrintInputFieldset(outf,sid,form,dn,sub_schema,required_attrs_dict,displayed_attrs,'Required attributes',input_form_entry)
    PrintInputFieldset(outf,sid,form,dn,sub_schema,allowed_attrs_dict,displayed_attrs,'Allowed attributes',input_form_entry)
    WriteInputFormFooter(outf,form,sub_schema,entry,required_attrs_dict,allowed_attrs_dict,input_formtype)


  elif input_formtype=='LDIF':
    WriteLDIFField(outf,form,ls,sub_schema,dn,entry,required_attrs_dict,allowed_attrs_dict)
    displayed_attrs = input_form_entry.keys()

  outf.write('</form></div>')
  w2lapp.gui.PrintFooter(outf,form)


def w2l_ModifyForm(sid,outf,command,form,ls,dn,Msg='',entry=None,skip_oc_input=0):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  if entry is None:
    entry = ReadOldEntry(ls,dn,sub_schema)

  if 'ldap_oc' in form.inputFieldNames:
    # Read objectclass(es) from input form
    entry['objectClass'] = [
      oc.encode('ascii')
      for oc in form.field['ldap_oc'].value
    ]
  elif not skip_oc_input:
    # Output the web page with object class input form
    objectClass,structuralObjectClass = ls.getObjectClasses(dn)
    if structuralObjectClass is None:
      structuralObjectClass = sub_schema.get_structural_oc(objectClass)
    ObjectClassForm(
      sid,outf,form,command,ls,sub_schema,dn,
      None,objectClass,structuralObjectClass
    )
    return

  required_attrs_dict,allowed_attrs_dict = entry.attribute_types()
  # Determine attributes which does not appear in the schema but
  # do exist in the entry
  nomatching_attrs_dict = ldap.cidict.cidict()
  for at_name in entry.keys():
    try:
      at_oid = sub_schema.name2oid[ldap.schema.AttributeType][at_name]
    except KeyError:
      nomatching_attrs_dict[at_name] = None
    else:
      if not (
        allowed_attrs_dict.has_key(at_oid) or \
        required_attrs_dict.has_key(at_oid) or \
        at_name.lower()=='objectclass'
      ):
        nomatching_attrs_dict[at_oid] = sub_schema.get_obj(ldap.schema.AttributeType,at_oid)

  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,
    heading_msg[command],
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

  input_form_entry = InputFormEntry(sid,form,ls,dn,sub_schema,entry)
  
  input_formtype = form.getInputValue('input_formtype',['Template'])[0]
  if not input_formtype in ['Template','Table','LDIF']:
    input_formtype = form.getInputValue('input_currentformtype',['Template'])[0]

  # Check whether to fall back to table input mode
  if input_formtype=='Template':
    template_oc,read_template_dict = w2lapp.read.DetermineHTMLTemplates(ls,entry,'input_template')
    if not template_oc:
      Msg = '<br>'.join((Msg,'<strong>No templates defined for chosen object classes.</strong>'))
      input_formtype='Table'

  outf.write("""<div id="Input" class="Main">
    <h1>%s</h1>
    %s
    %s
    <input type="hidden" name="dn" value="%s">
    <input type="submit" value="Modify">
    Change input form:
    <input type="submit" name="input_formtype" value="%s">
    <input type="submit" name="input_formtype" value="%s">
    %s
    <br>
    """ % (
      heading_msg[command],
      Msg,
      form.beginFormHTML('modify',sid,'POST',target=None,enctype='multipart/form-data'),
      dn.encode(form.accept_charset),
      {'Template':'Table','Table':'Template','LDIF':'Template'}[input_formtype],
      {'Template':'LDIF','Table':'LDIF','LDIF':'Table'}[input_formtype],
      ''.join([
        w2lapp.gui.HIDDEN_FIELD % ('in_oldattrtypes',at_name,'')
        for at_name in entry.keys()
      ]),
    )
  )

  # Use HTML templates for displaying a partial input form
  if input_formtype=='Template':
    displayed_attrs = w2lapp.read.PrintTemplateOutput(sid,outf,form,ls,sub_schema,entry,input_form_entry,'input_template')
    # Output all existing attributes as hidden fields which were not yet displayed
    # through HTML templates
    outf_lines = []
    for attr_type,attr_values in entry.items():
      at_oid = entry._at2key(attr_type)[0]
      if not at_oid in displayed_attrs:
        syntax_class = w2lapp.schema.syntaxes.syntax_registry.syntaxClass(sub_schema,attr_type)
        for attr_value in attr_values:
          if syntax_class.editable:
            attr_inst = syntax_class(sid,form,ls,dn,sub_schema,attr_type,attr_value,entry)
            outf_lines.append(w2lapp.gui.HIDDEN_FIELD % (
              'in_attrtype',utf2display(form.accept_charset,attr_type),''
            ))
            outf_lines.append(w2lapp.gui.HIDDEN_FIELD % (
              'in_value',utf2display(form.accept_charset,attr_inst.formValue()),''
            ))
    outf.write(''.join(outf_lines))

  elif input_formtype=='Table':
    # Displaying rest of input form as simple list
    displayed_attrs = {}
    PrintInputFieldset(outf,sid,form,dn,sub_schema,required_attrs_dict,displayed_attrs,'Required attributes',input_form_entry)
    PrintInputFieldset(outf,sid,form,dn,sub_schema,allowed_attrs_dict,displayed_attrs,'Allowed attributes',input_form_entry)
    if nomatching_attrs_dict:
      PrintInputFieldset(outf,sid,form,dn,sub_schema,nomatching_attrs_dict,displayed_attrs,'Existing attributes not matching schema',input_form_entry)
    WriteInputFormFooter(outf,form,sub_schema,entry,required_attrs_dict,allowed_attrs_dict,input_formtype)


  elif input_formtype=='LDIF':
    WriteLDIFField(outf,form,ls,sub_schema,dn,entry,required_attrs_dict,allowed_attrs_dict)
    displayed_attrs = input_form_entry.keys()

  assertion_filter = AssertionFilter(ls,entry)

  outf.write("""
  %s
  </form>
  </div>
  """ % (
    w2lapp.gui.HIDDEN_FIELD % (
      'in_assertion',
      utf2display(form.accept_charset,assertion_filter),
      '')
  ))
  w2lapp.gui.PrintFooter(outf,form)
