"""
w2lapp.form.py: class for web2ldap input form handling

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: form.py,v 1.68 2009/08/06 14:00:34 michael Exp $
"""

import urllib,re,ldif,ldapurl,ldap.schema,msgzip,ldaputil.base, \
       pyweblib.forms,ldapsession
       
import w2lapp.cnf,w2lapp.core,w2lapp.gui,w2lapp.passwd,w2lapp.searchform

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO


class Web2LDAPUrl(ldapurl.LDAPUrl):
  """
  Class for LDAP URLs passed as query string derived from LDAPUrl
  """
  attr2extype = {
    'who':'bindname',
    'cred':'X-BINDPW',
    'x_startTLS':'x-starttls',
    'saslMech':'x-saslmech',
    'saslAuthzId':'x-saslauthzid',
    'saslRealm':'x-saslrealm',
  }

  def getStartTLSOpt(self,minStartTLSOpt):
    """
    Returns a value indicating whether StartTLS ext.op. shall be used.
    Argument minStartTLSOpt indicates the minimum security level requested.
    0 No
    1 Yes, if possible. Proceed if not possible.
    2 Yes, mandantory. Abort if not possible.
    """
    try:
      e = self.extensions.get('startTLS',self.extensions['starttls'])
    except KeyError:
      try:
        result = int(self.x_startTLS or '0')
      except ValueError:
        raise w2lapp.core.ErrorExit('LDAP URL extension x-starttls must be integer 0, 1 or 2.')
    else:
      result = int(e.critical) + int(e.extype.lower()=='starttls')
    return max(result,minStartTLSOpt) # getStartTLSOpt()


class Web2LDAPForm(pyweblib.forms.Form):

  def __init__(self,inf,env,script_name=None):
    pyweblib.forms.Form.__init__(self,inf,env)
    self.script_name = script_name or self.script_name
    if env.has_key('HTTP_USER_AGENT'):
      self.browser_type,self.browser_version = pyweblib.helper.BrowserType(env['HTTP_USER_AGENT'])
    else:
      self.browser_type,self.browser_version = None,None
    self.determineCharset()
    self.initializeForm()

  def outFileObject(self,outf):

    """
    Create file object for gziping on-the-fly if configuration
    and HTTP header allows it
    """
    outfile = outf
    if w2lapp.cnf.misc.gzip_level>0 and \
       not msgzip.GzipFile is None and \
       self.http_accept_encoding['gzip']>0:
      # Create new gzip-stream file object
      return msgzip.GzipFile(
        filename='web2ldap stream output',
        mode='wb',
        compresslevel=w2lapp.cnf.misc.gzip_level,
        fileobj=outfile
      )
    else:
      return outfile

  def determineCharset(self):
    if w2lapp.cnf.misc.print_rawutf8==2:
      self.accept_charset = 'utf-8'
    elif w2lapp.cnf.misc.print_rawutf8==1:
      if self.http_accept_charset['utf-8']:
        self.accept_charset = 'utf-8'
      else:
        self.accept_charset = self.http_accept_charset.preferred()
    else:
      self.accept_charset = 'iso-8859-1'
    return # determineCharset()

  def initializeForm(self):
    """
    Add the required fields 
    """
    self.addGeneralFields()
    self.addCommandFields()

  def addGeneralFields(self):
    self.addField(pyweblib.forms.Input('who',u'Bind DN/AuthcID',1000,1,u'.*',size=40))
    self.addField(pyweblib.forms.Input('cred',u'with Password',200,1,u'.*',size=15))
    self.addField(pyweblib.forms.Input('login_authzid',u'SASL AuthzID',1000,1,u'.*',size=20))
    self.addField(pyweblib.forms.Input('login_realm',u'SASL Realm',1000,1,u'.*',size=20))
    self.addField(AuthMechSelect('login_mech',u'Authentication mechanism'))
    self.addField(pyweblib.forms.Input('ldapurl',u'LDAP Url',1024,1,'[ ]*ldap(|i|s)://.*',size=30))
    self.addField(pyweblib.forms.Input('host',u'Host:Port',255,1,w2lapp.gui.host_pattern,size=30))
    self.addField(DistinguishedNameInput('dn','Distinguished Name'))
    self.addField(pyweblib.forms.Select('scope','Scope',1,options=[('0','Base'),('1','One level'),('2','Sub tree')],default='2'))
    self.addField(DistinguishedNameInput('login_search_root','Login search root'))
    self.addField(pyweblib.forms.Input('login_filterstr',u'Login search filter string',300,1,'.*'))

  def addCommandFields(self):
    pass

  def actionUrlHTML(self,command,sid):
    return '%s/%s%s' % (
      self.script_name,
      command,
      {0:'/%s' % sid,1:''}[sid is None],
    )

  def beginFormHTML(self,command,sid,method,target=None,enctype=None):
    target = {0:'target="%s"' % (target),1:''}[target is None]
    return """
      <form
        action="%s"
        method="%s"
        %s
        enctype="%s"
        accept-charset="%s"
      >
      """  % (
        self.actionUrlHTML(command,sid),
        method,target,
        enctype or 'application/x-www-form-urlencoded',
        self.accept_charset
      )

  def formHTML(
    self,command,submitstr,sid,method,form_parameters,
    extrastr='',
    target=None
  ):
    """
    Build the HTML text of a submit form
    """
    utf2display = w2lapp.core.utf2display
    HIDDEN_FIELD = w2lapp.gui.HIDDEN_FIELD
    form_str = [
      self.beginFormHTML(command,sid,method,target)
    ]
    for param_name,param_value in form_parameters:
      form_str.append(HIDDEN_FIELD % (
          param_name,
          utf2display(self.accept_charset,param_value),
          ''
        )
      )
    form_str.append("""
      <input type="submit" value="%s">
      %s
      </form>""" % (submitstr,extrastr)
    )
    return '\n'.join(form_str)

  def applAnchor(
    self,command,anchor_text,sid,
    form_parameters,
    target=None,
    title=None
  ):
    """
    Build the HTML text of a anchor with form parameters
    """
    charset = self.accept_charset
    utf2display = w2lapp.core.utf2display
    target = {0:' target="%s"' % (target),1:''}[target is None]
    title = {0:' title="%s"' % (utf2display(charset,title)),1:''}[title is None]
    return '<a class="CommandLink"%s%s href="%s?%s">%s</a>' % (
      target,title,
      self.actionUrlHTML(command,sid),
      '&amp;'.join([
        '%s=%s' % (
          param_name,
          urllib.quote(param_value.encode(charset))
        )
        for param_name,param_value in form_parameters
      ]),
      anchor_text
    )


class Web2LDAPNullForm(Web2LDAPForm):
  pass


class Web2LDAPForm_searchform(Web2LDAPForm):

  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('searchform_mode',u'Search form mode',1,options=[(u'base',u'Base'),(u'adv',u'Advanced'),(u'exp',u'Expert')],default=u'base'))
    self.addField(pyweblib.forms.Field('search_params',u'Number of search parameters',2,1,'[0-9]+'))
    self.addField(DistinguishedNameInput('search_root','Search root'))
    self.addField(pyweblib.forms.Input('search_filterstr',u'Search filter string',400,1,'.*'))
    self.addField(pyweblib.forms.Input('searchform_template',u'Search form template name',60,w2lapp.cnf.misc.max_searchparams,u'[a-zA-Z0-9. ()-]+'))
    self.addField(
      pyweblib.forms.Select(
        'search_resnumber',
        u'Number of results to display',
        1,
        options=[(u'0',u'unlimited'),(u'10',u'10'),(u'20',u'20'),(u'50',u'50'),(u'100',u'100'),(u'200',u'200')],
        default='10'
      )
    )
    self.addField(pyweblib.forms.Input('search_attrs',u'Attributes to be read',1000,1,ur'[*+\w,_;-]+'))

class Web2LDAPForm_search(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('search_filterstr',u'Search filter string',400,1,u'.*'))
    self.addField(pyweblib.forms.Input('search_resminindex',u'Minimum index of search results',10,1,u'[0-9]+'))
    self.addField(pyweblib.forms.Input('search_resnumber',u'Number of results to display',3,1,u'[0-9]+'))
    self.addField(ExportFormatSelect('search_output'))
    self.addField(InclOpAttrsCheckbox('search_opattrs',u'Request operational attributes',default="yes",checked=0))
    self.addField(pyweblib.forms.Select('search_mode',u'Search Mode',1,options=[ur'(&%s)',ur'(|%s)']))
    self.addField(AttributeType('search_attr',u'Search attribute',w2lapp.cnf.misc.max_searchparams))
    self.addField(pyweblib.forms.Select('search_option',u'Search option',w2lapp.cnf.misc.max_searchparams,options=w2lapp.searchform.search_options))
    self.addField(pyweblib.forms.Input('search_string',u'Search string',60,w2lapp.cnf.misc.max_searchparams,u'.*'))
    self.addField(pyweblib.forms.Input('search_attrs',u'Attributes to be read',1000,1,ur'[*+\w,_;-]+'))

class Web2LDAPForm_modifyform(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('input_formtype',u'Type of input form',1,options=['Template','Table','LDIF'],default='Template'))
    self.addField(pyweblib.forms.Input('ldap_oc',u'Object classes',60,40,u'[a-zA-Z0-9.-]+'))

class Web2LDAPForm_addform(Web2LDAPForm_modifyform):
  def addCommandFields(self):
    Web2LDAPForm_modifyform.addCommandFields(self)
    self.addField(pyweblib.forms.Input('add_rdn','RDN of new entry',255,1,u'.*',size=50))
    self.addField(pyweblib.forms.Input('add_template',u'LDIF template name',60,w2lapp.cnf.misc.max_searchparams,u'[a-zA-Z0-9. ()-]+'))

class Web2LDAPForm_conninfo(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(
      pyweblib.forms.Radio(
        'ldap_managedsait',u'Manage DSA IT mode',maxValues=1,
        options=[(u'0',u'disabled'),(u'1',u'enabled')],default=u'0'
      )
    )
    self.addField(
      pyweblib.forms.Radio(
        'ldap_relaxrules',u'Relax rules',maxValues=1,
        options=[(u'0',u'disabled'),(u'1',u'enabled')],default=u'0'
      )
    )
    self.addField(
      pyweblib.forms.Select(
        'ldap_subentries',u'Subentries control',maxValues=1,
        options=[
          ('',u'off'),
          (ldapsession.CONTROL_SUBENTRIES,u'RFC 3672'),
          (ldapsession.CONTROL_LDUP_SUBENTRIES,u'draft-ietf-ldup-subentry')
        ],
        default=u'0'
      )
    )
    self.addField(
      pyweblib.forms.Radio(
        'ldap_dontusecopy',u"Don't use copy",maxValues=1,
        options=[(u'0',u'disabled'),(u'1',u'enabled')],default=u'0'
      )
    )
    self.addField(
      pyweblib.forms.Radio(
        'ldap_donotreplicate',u'Do not replicate',maxValues=1,
        options=[(u'0',u'disabled'),(u'1',u'enabled')],default=u'0'
      )
    )
    self.addField(
      pyweblib.forms.Radio(
        'ldap_serveradministration',u'Server administration',maxValues=1,
        options=[(u'0',u'disabled'),(u'1',u'enabled')],default=u'0'
      )
    )
    self.addField(pyweblib.forms.Select('conninfo_flushcaches',u'Flush caches',1,options=['0','1'],default=0))
    self.addField(
      pyweblib.forms.Select(
        'ldap_deref',u'Dereference aliases',maxValues=1,default=str(ldap.DEREF_NEVER),
        options=[
          (str(ldap.DEREF_NEVER),u'never'),
          (str(ldap.DEREF_SEARCHING),u'searching'),
          (str(ldap.DEREF_FINDING),u'finding'),
          (str(ldap.DEREF_ALWAYS),u'always'),
        ]
      )
    )

class Web2LDAPForm_input(Web2LDAPForm):
  """Base class for entry data input not directly used"""
  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('input_formtype',u'Type of input form',1,options=['Template','Table','LDIF','+'],default='Template'))
    self.addField(pyweblib.forms.Select('input_currentformtype',u'Type of input form',1,options=['Template','Table','LDIF'],default='Template'))
    self.addField(AttributeType('in_attrtype',u'Attribute type',w2lapp.cnf.misc.input_maxattrs))
    self.addField(AttributeType('in_binattrtype',u'Binary attribute type',w2lapp.cnf.misc.input_maxbinattrs))
    self.addField(AttributeType('input_addattrtype',u'Attribute type to be extended',1))
    self.addField(
      pyweblib.forms.Input(
        'in_value',u'Attribute Value',
        w2lapp.cnf.misc.input_maxfieldlen,
        w2lapp.cnf.misc.input_maxattrs,
        ('.*',re.U|re.M|re.S)
      )
    )
    self.addField(
      pyweblib.forms.File(
        'in_binvalue',u'Binary attribute value',
        w2lapp.cnf.misc.input_maxfilelen,
        w2lapp.cnf.misc.input_maxattrs,
        None
      )
    )
    self.addField(LDIFTextArea('in_ldif',u'LDIF data'))

class Web2LDAPForm_add(Web2LDAPForm_input):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('add_rdn',u'RDN of new entry',255,1,u'.*',size=50))
    Web2LDAPForm_input.addCommandFields(self)

class Web2LDAPForm_modify(Web2LDAPForm_input):
  def addCommandFields(self):
    self.addField(AttributeType('in_oldattrtypes',u'Old attribute types',w2lapp.cnf.misc.input_maxattrs))
    self.addField(pyweblib.forms.Input('in_assertion',u'Assertion filter string',2000,1,'.*',required=1))
    Web2LDAPForm_input.addCommandFields(self)

class Web2LDAPForm_delete(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('delete_confirm',u'Confirmation',1,options=['yes','no'],default='no'))
    self.addField(pyweblib.forms.Checkbox('delete_usecontrol',u'Use tree delete control',1,default="yes",checked=0))
    self.addField(pyweblib.forms.Input('delete_attr',u'Attribute to be deleted',255,100,u'[\w_;-]+'))

class Web2LDAPForm_rename(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('rename_newrdn',u'New RDN',255,1,ldaputil.base.rdn_pattern,size=50))
    self.addField(DistinguishedNameInput('rename_newsuperior','New superior DN'))
    self.addField(pyweblib.forms.Checkbox('rename_delold',u'Delete old',1,default="yes",checked=1))
    self.addField(
      pyweblib.forms.Input(
        'rename_newsupfilter',
        u'Filter string for searching new superior entry',300,1,'.*',
        default='(|(objectClass=organization)(objectClass=organizationalUnit))',
        size=50,
      )
    )
    self.addField(DistinguishedNameInput('rename_searchroot','Search root under which to look for new superior entry.'))
    self.addField(pyweblib.forms.Input('rename_supsearchurl',u'LDAP URL for searching new superior entry',100,1,'.*',size=30))

class Web2LDAPForm_passwd(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(
      pyweblib.forms.Select(
        'passwd_action',
        u'Password action',
        1,
        options=[ (action,short_desc) for action,short_desc,long_desc in w2lapp.passwd.PASSWD_ACTIONS ],
        default='setuserpassword'
    ))
    self.addField(DistinguishedNameInput('passwd_who',u'Password DN'))
    self.addField(pyweblib.forms.Field('passwd_newpasswd',u'New password',100,2,'.*'))
    self.addField(pyweblib.forms.Select('passwd_scheme',u'Password hash scheme',1,options=w2lapp.passwd.available_hashtypes,default=w2lapp.passwd.available_hashtypes[-1]))
    self.addField(pyweblib.forms.Checkbox('passwd_ntpasswordsync',u'Sync ntPassword for Samba',1,default="yes",checked=1))
    self.addField(pyweblib.forms.Checkbox('passwd_lmpasswordsync',u'Sync lmPassword for Samba',1,default="yes",checked=1))
    self.addField(pyweblib.forms.Checkbox('passwd_settimesync',u'Sync password setting times',1,default="yes",checked=1))

class Web2LDAPForm_read(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('read_nocache',u'Force fresh read',1,options=['0','1'],default=0))
    self.addField(pyweblib.forms.Input('read_attr',u'Read attribute',255,100,u'[\w_;-]+'))
    self.addField(pyweblib.forms.Select('read_attrmode',u'Read attribute',1,options=[u'view',u'load']))
    self.addField(pyweblib.forms.Input('read_attrindex',u'Read attribute',255,1,u'[0-9]+'))
    self.addField(pyweblib.forms.Input('read_attrmimetype',u'MIME type',255,1,u'[\w.-]+/[\w.-]+'))
    self.addField(pyweblib.forms.Select('read_output',u'Read output format',1,options=['table','vcard','template'],default='template'))
    self.addField(pyweblib.forms.Input('read_filterstr',u'Search filter string when reading single entry',400,1,'.*'))

class Web2LDAPForm_groupadm(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(DistinguishedNameInput('group_search_root','Group search root'))
    self.addField(DistinguishedNameInput('groupadm_add','Add to group',300))
    self.addField(DistinguishedNameInput('groupadm_remove','Remove from group',300))
    self.addField(pyweblib.forms.Select('groupadm_view',u'Group list view',1,options=[('0','none of the'),('1','only member'),('2','all')],default='1'))

class Web2LDAPForm_login(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(DistinguishedNameInput('login_who',u'Bind DN'))

class Web2LDAPForm_locate(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(
      pyweblib.forms.Input(
        'locate_name',u'Location name',
        255,1,u'(%s|%s|%s)' % (
          w2lapp.schema.syntaxes.mail_pattern[1:-1],
          w2lapp.gui.host_pattern,
          ldaputil.base.dn_pattern
        ),
        size=25
    ))

class Web2LDAPForm_oid(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(OIDInput('oid',u'OID'))
    self.addField(pyweblib.forms.Select('oid_class','Schema element class',1,options=ldap.schema.SCHEMA_ATTRS,default=''))



FORM_CLASS = {
  '':Web2LDAPForm,
  'monitor':Web2LDAPNullForm,
  'urlredirect':Web2LDAPNullForm,
  'disconnect':Web2LDAPNullForm,
}

_FORM_CLASS_NAME_PREFIX = 'Web2LDAPForm_'
_COMMAND_STR_OFFSET = len(_FORM_CLASS_NAME_PREFIX)

for _name in dir():
  if _name.startswith(_FORM_CLASS_NAME_PREFIX):
    c = eval(_name)
    try:
      command = _name[_COMMAND_STR_OFFSET:]
    except IndexError:
      pass
    else:
      FORM_CLASS[command] = c
      del command



class DistinguishedNameInput(pyweblib.forms.Input):
  """Input field class for LDAP DNs."""

  def __init__(self,name='dn',text='DN',maxValues=1,required=0,default=''):
    pyweblib.forms.Input.__init__(
      self,name,text,1024,maxValues,'',
      size=50,required=required,default=default
    )

  def _validateFormat(self,value):
    if value and not ldaputil.base.is_dn(value):
      raise pyweblib.forms.InvalidValueFormat(
        self.name,
        self.text.encode(self.charset),
        value.encode(self.charset)
      )


class LDIFTextArea(pyweblib.forms.Textarea):
  """A single multi-line input field for LDIF data"""

  def __init__(
    self,name='in_ldif',text='LDIF data',required=0,max_entries=1
  ):
    pyweblib.forms.Textarea.__init__(
      self,
      name,text,w2lapp.cnf.misc.ldif_maxbytes,1,ldif.ldif_pattern,
      required=required,
    )
    self._max_entries = max_entries
    self.allRecords = []

  def getLDIFRecords(self):
    if self.value:
      f = StringIO('\n'.join(self.value).encode(self.charset))
      ldif_parser = ldif.LDIFRecordList(
        f,
        ignored_attr_types=[],
        max_entries=self._max_entries,
        process_url_schemes=w2lapp.cnf.misc.ldif_url_schemes
      )
      # It's not clear why ldif_parser stays persistent.
      # But let's reset class attribute now.
      # FIX ME!!! Likely not thread-safe!
      ldif_parser.all_records = []
      ldif_parser.parse()
      return ldif_parser.all_records[:]
    else:
      return []


class OIDInput(pyweblib.forms.Input):

  def __init__(self,name,text,default=None):
    pyweblib.forms.Input.__init__(
      self,name,text,
      512,1,u'[a-zA-Z0-9_.;*-]+',
      default=default,
      required=0,size=25
    )


class ObjectClassSelect(pyweblib.forms.Select):
  """Select field class for choosing the object class(es)"""
  def __init__(
    self,
    name='ldap_oc',
    text='Object classes',
    options=None,
    default=None,
    required=0,
    accesskey='',
    size=12,            # Size of displayed select field
  ):
    select_default = default or []
    select_default.sort()
    additional_options = [ i for i in options or [] if not i in select_default ]
    additional_options.sort()
    select_options = select_default[:]
    select_options.extend(additional_options)
    pyweblib.forms.Select.__init__(
      self,
      name,text,maxValues=200,
      required=required,
      options=select_options,
      default=select_default,
      accessKey=accesskey,
      size=size,
      ignoreCase=1,
      multiSelect=1
    )
    self.setRegex('[\w]+')
    self.maxLen = 200
  

class ExportFormatSelect(pyweblib.forms.Select):
  """Select field class for choosing export format"""

  def __init__(
    self,
    name='search_output',
    text=u'Export format',
    options=None,
    default='ldif1',
    required=0,
  ):
    pyweblib.forms.Select.__init__(
      self,
      name,text,1,
      options=options or [
        (u'table',u'Table/template'),
        (u'raw',u'Raw DN list'),
        (u'print',u'Printable'),
        (u'ldif',u'LDIF (Umich)'),
        (u'ldif1',u'LDIFv1 (RFC2849)'),
        (u'dsml1',u'DSMLv1'),
      ],
      default=default,
      required=required,
      size=1,
    )


class AttributeType(pyweblib.forms.Input):
 
  def __init__(self,name,text,maxValues):
    pyweblib.forms.Input.__init__(
      self,name,text,500,maxValues,
      ldaputil.base.attr_type_pattern,required=0,size=30
    )


class InclOpAttrsCheckbox(pyweblib.forms.Checkbox):

  def __init__(self,name,text,default='yes',checked=0):
    pyweblib.forms.Checkbox.__init__(self,name,text,1,default=default,checked=checked)


class AuthMechSelect(pyweblib.forms.Select):
  """Select field class for choosing the bind mech"""

  supported_bind_mechs = {
    '':'Simple Bind',
    'DIGEST-MD5':'SASL Bind: DIGEST-MD5',
    'CRAM-MD5':'SASL Bind: CRAM-MD5',
    'PLAIN':'SASL Bind: PLAIN',
    'LOGIN':'SASL Bind: LOGIN',
    'GSSAPI':'SASL Bind: GSSAPI',
    'EXTERNAL':'SASL Bind: EXTERNAL',
    'OTP':'SASL Bind: OTP',
    'NTLM':'SASL Bind: NTLM',
  }

  def __init__(
    self,
    name='login_mech',
    text=u'Authentication mechanism',
    default=None,
    required=0,
    accesskey='',
    size=1,            # Size of displayed select field
  ):
    pyweblib.forms.Select.__init__(
      self,
      name,text,maxValues=1,
      required=required,
      options=None,
      default=default or [],
      accessKey=accesskey,
      size=size,
      ignoreCase=0,
      multiSelect=0
    )

  def setOptions(self,options):
    options_dict = {}
    options_dict[''] = self.supported_bind_mechs['']
    for o in options or self.supported_bind_mechs.keys():
      o_upper = o.upper()
      if self.supported_bind_mechs.has_key(o_upper):
        options_dict[o_upper] = self.supported_bind_mechs[o_upper]
    pyweblib.forms.Select.setOptions(self,options_dict.items())
