"""
w2lapp.gui.py: basic functions for GUI elements

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: gui.py,v 1.200 2009/08/06 14:00:34 michael Exp $
"""

import os,urllib,ldap, \
       ldaputil,pyweblib.forms,pyweblib.httphelper,msbase, \
       w2lapp.core,w2lapp.cnf,w2lapp.schema.syntaxes

from ldapurl import LDAPUrl,LDAPUrlExtension,isLDAPUrl
from pyweblib.forms import escapeHTML

from ldaputil.base import ParentDN,ParentDNList,explode_dn

from w2lapp.core import utf2display

from types import UnicodeType

########################################################################
# Initialize some constants used throughout web2ldap
########################################################################

host_pattern = '[a-zA-Z0-9_.:\[\]-]+'

HIDDEN_FIELD = '<input type="hidden" name="%s" value="%s">%s\n'

# This function searches for variants
def GetVariantFilename(pathname,variantlist):
  if not pathname:
    raise w2lapp.core.ErrorExit('No template file defined.')
  for v in variantlist:
    variant_filename = pathname+'.'+v
    if os.path.isfile(variant_filename):
      break
  else:
    variant_filename = pathname
  return variant_filename


command_method = {
  'add':'POST',
  'modify':'POST',
  'rename':'POST',
  'login':'POST',
  'passwd':'POST',
  'delete':'POST',
}


def LDAPError2ErrMsg(e):
  """
  Converts a LDAPError exception into HTML error message
  """
  if isinstance(e,ldap.TIMEOUT) or not e.args:
    ErrMsg = ''
  else:
    try:
      ErrMsg = '%s: %s' % (e.args[0]['desc'],e.args[0].get('info',''))
    except TypeError:
      try:
        ErrMsg = '%s: %s' % (e[0],e[1])
      except TypeError,IndexError:
        ErrMsg = str(e)
  return ErrMsg


def ConnLDAPUrl(ls,dn):
  lu = LDAPUrl(ldapUrl=ls.uri.encode('ascii'))
  lu.dn = dn.encode(ls.charset)
  lu.who=(ls.who or u'').encode(ls.charset) or None
  return lu # ConnLDAPUrl()


def DisplayDN(sid,form,ls,dn,commandbutton=0,active_ldapurl=0):
  """Display a DN as LDAP URL with or without button"""
  assert type(dn)==UnicodeType, "Argument 'dn' must be UnicodeType"
  dn_str = utf2display(form.accept_charset,dn or u'- World -')
  if active_ldapurl:
    href_str = ConnLDAPUrl(ls,dn).htmlHREF(hrefText=dn_str,hrefTarget=None)
  else:
    href_str = dn_str
  if commandbutton:
    return ' | '.join([
      href_str,
      form.applAnchor('read','Read',sid,[('dn',dn)])
    ])
  else:
    return href_str


def CommandTable(
  outf,
  commandlist,
  div_id='CommandDiv',
  separator=' | '
):
  if commandlist:
    outf.write('<p id="%s" class="CommandTable">\n%s\n</p>\n' % (
      div_id,
      (separator).join(commandlist)
    ))
  return # CommandTable()


def ContextMenuSingleEntry(sid,form,ls,dn,vcard_link=0):
  """Output the context menu for a single entry"""
  result = [
    form.applAnchor('read','Raw',sid,[('dn',dn),('read_output','table')],title='Display this entry as raw attribute type/value list'),
    form.applAnchor('login','Bind as',sid,[('dn',dn),('login_who',dn)],title="Login with this entry's DN as bind-DN"),
    form.applAnchor('modifyform','Modify',sid,[('dn',dn)],title='Modify this entry'),
    form.applAnchor('rename','Rename',sid,[('dn',dn)],title='Rename/move this entry'),
    form.applAnchor('delete','Delete',sid,[('dn',dn)],title='Delete this entry and/or subtree'),
    form.applAnchor('passwd','Password',sid,[('dn',dn),('passwd_who',dn)],title='Set user password for this entry'),
    form.applAnchor('groupadm','Groups',sid,[('dn',dn)],title='Change group membership for this entry'),
  ]
  if vcard_link:
    result.append(form.applAnchor('read','vCard',sid,[('dn',dn),('read_output','vcard')],title='Export this entry as vCard'))
  if not ls.currentAuditContext is None:
    result.extend([
      form.applAnchor(
        'search','Audit access',sid,
        [
          ('dn',ls.currentAuditContext.encode(ls.charset)),
          ('search_filterstr','(&(objectClass=auditObject)(reqDN=%s))' % (ldap.filter.escape_filter_chars(dn))),
          ('scope',str(ldap.SCOPE_ONELEVEL)),
        ],
        title='Complete audit trail for current entry',
      ),
      form.applAnchor(
        'search','Audit writes',sid,
        [
          ('dn',ls.currentAuditContext.encode(ls.charset)),
          ('search_filterstr','(&(objectClass=auditWriteObject)(reqDN=%s))' % (ldap.filter.escape_filter_chars(dn))),
          ('scope',str(ldap.SCOPE_ONELEVEL)),
        ],
        title='Audit trail of write access to current entry',
      ),
    ])
  return result


def StatusLine(sid,outf,form,ls,dn):
  """
  Status line displaying host, current DN, LDAP URL and bind DN
  """
  template_dict = msbase.CaseinsensitiveStringKeyDict(default='&nbsp;')
  template_dict.update(form.env)
  template_dict.update({
    'sid':sid or '',
    'SCRIPT_NAME':form.script_name,
    'web2ldap_version':escapeHTML(w2lapp.__version__),
  })
  if ls!=None and ls.uri!=None:
    # Only output something meaningful if valid connection
    template_dict.update({
      'host':ConnLDAPUrl(ls,dn).htmlHREF(hrefText=(ls.uri or u'disconnected host').encode(form.accept_charset),hrefTarget=None),
#      'dn':DisplayDN(sid,form,ls,dn,active_ldapurl=1),
      'description':w2lapp.cnf.GetParam(ls,'description',u'').encode(form.accept_charset),
      'dit_navi':','.join(DITNavigationList(sid,outf,form,ls,dn,level=0)),
    })
    if hasattr(ls,'who') and ls.who:
      template_dict['who'] = DisplayDN(sid,form,ls,ls.who)
    else:
      template_dict['who'] = 'anonymous'
  else:
    template_dict.update({
      'host':'not connected',
    })
  status_line_template = """<table id="StatusLine" summary="LDAP connection status information">
    <tr>
      <td>
        Server:
      </td>
      <td>
        %(host)s: &quot;%(description)s&quot;
      </td>
      <td rowspan="3" id="About">
        <a href="%(SCRIPT_NAME)s/urlredirect/%(sid)s?http://www.web2ldap.de/">
          web2ldap %(web2ldap_version)s
        </a>
      </td>
    </tr>
    <tr>
      <td>
        Bound as:
      </td>
      <td>
        %(who)s
      </td>
    </tr>
    <tr>
      <td>
        Browse DIT:
      </td>
      <td>
        %(dit_navi)s
      </td>
    </tr>
  </table>
  """
  outf.write(status_line_template % template_dict)


def MainMenu(sid,form,ls,dn):
  """
  Returns list of main menu items
  """
  cl = []

  if ls!=None:

    if ls.uri!=None:

      if dn:
        cl.append(
          form.applAnchor(
            'search','Up',sid,
            [
              ('dn',ParentDN(dn)),
              ('scope',str(ldap.SCOPE_ONELEVEL)),
            ],
            title='Go up one level',
          )
        )
      cl.append(
        form.applAnchor(
          'search','Down',sid,
          [
            ('dn',dn),
            ('scope',str(ldap.SCOPE_ONELEVEL)),
          ],
          title='Descend into tree from here',
        )
      )
      
      cl.append(
        form.applAnchor('searchform','Search',sid,[('dn',dn)],
          title='Enter search criteria in input form',
        ),
      )
      if dn or ls.l.get_option(ldap.OPT_PROTOCOL_VERSION)>=ldap.VERSION3:
        cl.append(
          form.applAnchor(
            'read',
            {0:'Read',1:'[Root DSE]'}[dn==''],
            sid,[('dn',dn),('read_nocache','1')],
            title='Read the current entry',
          ),
        )
      cl.append(
        form.applAnchor('addform','New entry',sid,[('dn',dn)],
        title='Add a new entry below here',
      ))
      cl.append(
        form.applAnchor('conninfo','ConnInfo',sid,[('dn',dn)],
          title='Show information about HTTP and LDAP connections',
      ))
      cl.append(
        form.applAnchor('login','Bind',sid,[('dn',dn)],
          title='Login to directory',
      ))

  cl.append(form.applAnchor('disconnect','Disconnect',sid,[],title='Disconnect from LDAP server'))
  
  return cl # MainMenu()


def DITNavigationList(sid,outf,form,ls,dn,level=0):
  dn_list=explode_dn(dn)[level:]
  dn_list_len = len(dn_list)
  result = [
    form.applAnchor(
      'read',
      utf2display(form.accept_charset,dn_list[i] or '[Root DSE]'),
      sid,
      [
        ('dn',','.join(dn_list[i:dn_list_len])),
#        ('scope',str(ldap.SCOPE_ONELEVEL)),
      ],
      title=u'Jump to %s' % (u','.join(dn_list[i:dn_list_len])),
    )
    for i in range(dn_list_len)
  ]
#  if dn:
  result.append(
    form.applAnchor(
      'read',
      '[Root DSE]',
      sid,
      [
        ('dn',''),
#        ('scope',str(ldap.SCOPE_ONELEVEL)),
      ],
      title='Jump to root DSE',
    )
  )
  return result # DITNavigationList()


def TopSection(sid,outf,form,ls,dn,title,main_menu_list,context_menu_list=[]):
  if not ldaputil.base.is_dn(dn):
    dn=''
  web2ldap_title = ' - '.join(filter(None,[
    utf2display(form.accept_charset,dn),title
  ]))
  w2lapp.gui.PrintHeader(sid,outf,form,ls,web2ldap_title)
  outf.write('<div id="TopSection">\n')
  StatusLine(sid,outf,form,ls,dn)
  CommandTable(outf,main_menu_list,div_id='MainMenu')
  CommandTable(outf,context_menu_list,div_id='ContextMenu')
  outf.write('</div>\n')
  return # TopSection()


def SimpleMessage(
  sid,outf,form,ls,dn,
  title='',message='',
  p_id='Message',
  main_menu_list=[],context_menu_list=[]
):
  TopSection(sid,outf,form,ls,dn,title,main_menu_list,context_menu_list=context_menu_list)
  outf.write('<div id="%s" class="Main">\n%s\n</div>\n' % (p_id,message))
  w2lapp.gui.PrintFooter(outf,form)
  return # SimpleMessage()


# Return a pretty HTML-formatted string describing a schema element
# referenced by name or OID
def SchemaElementName(sid,form,dn,schema,se_nameoroid,se_class,name_template=r'<strong>%s</strong>'):
  result = [name_template % (se_nameoroid.encode())]
  if se_class:
    se = schema.get_obj(se_class,se_nameoroid,None)
    if not se is None:
      result.append(form.applAnchor(
          'oid','&raquo;',sid,
          [ ('dn',dn),('oid',se.oid),('oid_class',ldap.schema.SCHEMA_ATTR_MAPPING[se_class]) ]
      ))
  return '\n'.join(result)


def LDAPURLButton(sid,form,ls,data):
  if isinstance(data,LDAPUrl):
    l = data
  else:
    l = LDAPUrl(ldapUrl=data)
  command_hidden_fields = []
  if l.scope!=None:
    command_hidden_fields.append(('scope',str(l.scope)))
  if l.filterstr!=None:
    command_hidden_fields.append(('search_filterstr',l.filterstr))
  if l.scope!=ldap.SCOPE_BASE:
    command_func,command_text = 'search','Search'
  else:
    command_func,command_text = 'read','Read'
  command_hidden_fields.extend([
    ('host',l.hostport),
    ('dn',l.dn),
  ])
  return form.applAnchor(command_func,command_text,sid,command_hidden_fields)


def MailToURL(mailadr):
  return '<a href="mailto:%s">%s</a>' % (urllib.quote(mailadr),mailadr)


def DataStr(sid,form,ls,dn,schema,attrtype_name,value,valueindex=0,commandbutton=0,entry=None):
  """
  Return a pretty HTML-formatted string of the attribute value
  """
  attr_instance = w2lapp.schema.syntaxes.syntax_registry.attrInstance(sid,form,ls,dn,schema,attrtype_name,value,entry)
  return attr_instance.displayValue(valueindex,commandbutton)


# Ausdrucken eines HTML-Kopfes mit Titelzeile
def PrintHeader(sid,outf,form,ls,TitleMsg):
  
  pyweblib.httphelper.SendHeader(
    outf,
    'text/html',
    form.accept_charset,
    expires_offset=w2lapp.cnf.misc.sec_expire,
    additional_header={
      'Pragma':'no-cache',
      'Cache-Control':'private,no-store,no-cache,max-age=0,must-revalidate'
    },
  )

  outf.write("""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
  "http://www.w3.org/TR/html4/loose.dtd">
<html>
  <head>
    <title>web2ldap - %s</title>
    <meta name="generator" content="web2ldap %s, see www.web2ldap.de">
    <meta http-equiv="Content-Type" content="text/html;charset=%s">
    <!--Begin of configured link_css-->
    %s
    <!--End of configured link_css-->
    <!--Begin of configured html_head-->
    %s
    <!--End of configured html_head-->
  </head>
%s
""" % (
    TitleMsg,
    w2lapp.__version__,
    form.accept_charset,
    w2lapp.cnf.GetParam(ls,'link_css',''),
    w2lapp.cnf.misc.html_head,
    w2lapp.cnf.misc.html_bodybegin
  ))
  return


# Ausdrucken eines HTML-Endes
def PrintFooter(f,form):
  f.write('</body>\n</html>\n')
  f.flush()


def SearchRootField(form,ls,dn,name='dn',text=u'Search Root',default=None):
  """Prepare input field for search root"""
  if dn:
    dn_select_list = [dn]+ParentDNList(dn,ls.getSearchRoot(dn))
  else:
    dn_select_list = []
  dn_select_list = msbase.union(ls.getNamingContexts(),dn_select_list,ignorecase=1)
  dn_select_list.append((u'',u'- World -'))
  srf = pyweblib.forms.Select(name,text,1,default=default,options=dn_select_list,ignoreCase=1)
  srf.setDefault(ls.getSearchRoot(dn))
  srf.setCharset(form.accept_charset)
  return srf


def ExceptionMsg(sid,outf,form,ls,dn,Heading,HTMLMsg):
  TopSection(sid,outf,form,ls,dn,'Error',MainMenu(sid,form,ls,dn),context_menu_list=[])
  outf.write("""
<div id="ErrorMessage" class="Main">
  <h1>%s</h1>
  <p>
    %s
  </p>
  <p>
    For questions or error reports contact:
    <a href="mailto:%s">%s</a>
  </p>
</div>
""" % (utf2display(form.accept_charset,Heading),HTMLMsg,form.env.get('SERVER_ADMIN',''),form.env.get('SERVER_ADMIN','')))
  PrintFooter(outf,form)

