"""
w2lapp.modify.py: modify an entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: modify.py,v 1.90 2009/04/02 17:53:49 michael Exp $
"""

import ldap,ldif,ldap.schema, \
       ldapsession,w2lapp.core,w2lapp.cnf,w2lapp.gui,w2lapp.addmodifyform
       
from w2lapp.schema.syntaxes import syntax_registry,LDAPSyntaxValueError

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO

from ldaputil.modlist2 import modifyModlist

def GetEntryfromInputForm(form,ls,dn,sub_schema):

  invalid_attrs = []

  e = w2lapp.addmodifyform.UserEditableEntry(ls,sub_schema,dn.encode(ls.charset),{})
  try:
    in_ldif = form.field['in_ldif'].getLDIFRecords()
  except ValueError,e:
    raise w2lapp.core.ErrorExit('LDIF parsing error: %s' % (str(e)))
  else:
    if in_ldif:
      e.update(in_ldif[0][1])

  # Get all the attribute types
  in_attrtype_list = [
    a.encode('ascii')
    for a in form.getInputValue('in_attrtype',[])
  ]
  in_binattrs = [
    a.encode('ascii')
    for a in form.getInputValue('in_binattrtype',[])
  ]
  in_attrtype_list.extend(in_binattrs)
  # Grab the Unicode input strings and convert to LDAP charset
  in_value_list = [
    a.encode(ls.charset)
    for a in form.getInputValue('in_value',[])
  ]
  # Grab the file-based input and leave as is
  in_value_list.extend([
    a for a in form.getInputValue('in_binvalue',[])
  ])
  
  if len(in_attrtype_list)!=len(in_value_list):
    raise w2lapp.core.ErrorExit('Different count of attribute types and values.')

  for i in range(len(in_attrtype_list)):
    attr_type = in_attrtype_list[i]
    syntax_class = syntax_registry.syntaxClass(sub_schema,attr_type)
    attr_instance = syntax_class(None,form,ls,dn,sub_schema,attr_type,None)
    try:
      attr_instance.setAttrValue(attr_instance.sanitizeInput(in_value_list[i]))
    except LDAPSyntaxValueError:
      invalid_attrs.append(attr_type)
    else:
      if attr_instance.attrValue:
        if not e.has_key(attr_type):
          e[attr_type] = []
        e[attr_type].append(attr_instance.attrValue)

  return e,invalid_attrs,in_binattrs # GetEntryfromInputForm()


def ModlistLDIF(dn,modlist):
  """
  Return a string containing a HTML table showing modifiy operations
  and attr type/value pairs
  """
  s = []
  s.append('<pre summary="Modify list">')
  f = StringIO()
  ldif_writer = ldif.LDIFWriter(f)
  ldif_writer.unparse(dn.encode('utf-8'),modlist)
  s.append(f.getvalue())
  s.append('</pre>')
  return '\n'.join(s) # ModlistTable()


##############################################################################
# Modify existing entry
##############################################################################

def w2l_Modify(sid,outf,command,form,ls,dn):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  new_entry,invalid_attrs,in_binattrs = GetEntryfromInputForm(form,ls,dn,sub_schema)

  # Check if the user just switched input type form
  if 'input_formtype' in form.inputFieldNames:
    input_addattrtypes = form.getInputValue('input_addattrtype',[])
    for input_addattrtype in input_addattrtypes:
      if input_addattrtype:
        new_entry[input_addattrtype.encode('ascii')] = new_entry.get(input_addattrtype,[])+['']
    w2lapp.addmodifyform.w2l_ModifyForm(
      sid,outf,'modifyform',form,ls,dn,
      Msg='',
      entry=new_entry,
      skip_oc_input=1
    )
    return

  if invalid_attrs:
    w2lapp.addmodifyform.w2l_ModifyForm(
      sid,outf,'modifyform',form,ls,dn,
      Msg='Syntax check failed for the following attributes: %s' % (
        ', '.join(invalid_attrs)
      ),
      entry=new_entry,
      skip_oc_input=1
    )
    return

  in_oldattrtypes = {}
  for a in form.getInputValue('in_oldattrtypes',[]):
    attr_type = a.encode('ascii')
    in_oldattrtypes[attr_type] = None

  in_assertion = form.getInputValue(
    'in_assertion',['(objectClass=*)']
  )[0]

  constant_attrs = w2lapp.addmodifyform.ConfiguredConstantAttributes(ls)

  try:
    old_entry = w2lapp.addmodifyform.ReadOldEntry(ls,dn,sub_schema,assertion_filter=in_assertion)
  except ldap.NO_SUCH_OBJECT:
    raise w2lapp.core.ErrorExit('Old entry was removed or modified!')

  # Set up a dictionary of all attribute types to be ignored
  ignore_attr_types = ldap.cidict.cidict()
  # Add all attributes which have NO-USER-MODIFICATION set
  ignore_attr_types.update(sub_schema.no_user_mod_attr_oids)
  # All attributes currently read which were not viewable before
  # must be ignored to avoid problems with different access rights
  ignore_attr_types.update({}.fromkeys([
    a
    for a in old_entry.keys()
    if not in_oldattrtypes.has_key(a)
  ]))
  # Ignore attributes which are assumed to be constant (some operational attributes)
  ignore_attr_types.update(constant_attrs)
  # Ignore binary attributes in new form input
  ignore_attr_types.update({}.fromkeys(in_binattrs))
  # Ignore binary attributes from old entry data
  for attr_type in old_entry.keys():
    syntax_class = syntax_registry.syntaxClass(sub_schema,attr_type)
    if not syntax_class.editable:
      ignore_attr_types[attr_type] = None

  relax_rules_enabled = ls.l._get_server_ctrls('__write__').has_key(ldapsession.CONTROL_RELAXRULES)
  if not relax_rules_enabled:
    required_attrs_dict,allowed_attrs_dict = new_entry.attribute_types()

  try:
    del ignore_attr_types['2.5.4.0']
  except KeyError:
    pass

  # Create modlist containing deltas
  modlist = modifyModlist(
    sub_schema,
    old_entry,new_entry,
    ignore_attr_types=ignore_attr_types.keys(),
    ignore_oldexistent=0,
  )
  # Binary values are always replaced
  for attr_type in new_entry.keys():
    syntax_class = syntax_registry.syntaxClass(sub_schema,attr_type)
    if (not syntax_class.editable) and new_entry[attr_type]:
      modlist.append((ldap.MOD_REPLACE,attr_type,new_entry[attr_type]))

  if modlist:
    # Send modify-list to host
    try:
      ls.modifyEntry(dn,modlist)
    except (
      ldap.CONSTRAINT_VIOLATION,
      ldap.INVALID_SYNTAX,
      ldap.OBJECT_CLASS_VIOLATION,
      ldap.OTHER,
      ldap.TYPE_OR_VALUE_EXISTS,
      ldap.UNDEFINED_TYPE,
    ),e:
      w2lapp.addmodifyform.w2l_ModifyForm(
        sid,outf,'modifyform',form,ls,dn,
        Msg=w2lapp.gui.LDAPError2ErrMsg(e),
        entry=new_entry,
        skip_oc_input=1
      )
      return
    else:
      # delete all cache entries referencing to old DN
      UserMsg = 'Modified entry %s with your input data:</p>\n%s' % (
        w2lapp.gui.DisplayDN(sid,form,ls,dn),
        ModlistLDIF(dn,modlist).encode(form.accept_charset)
      )
  else:
    UserMsg = 'No attributes modified of entry %s.' % (w2lapp.gui.DisplayDN(sid,form,ls,dn))

  # Output comes here
  w2lapp.gui.SimpleMessage(
    sid,outf,form,ls,dn,
    'Modify result',
    UserMsg,
    main_menu_list=w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

