"""
web2ldap plugin classes for Samba 3

$Id: samba.py,v 1.13 2009/08/06 14:00:34 michael Exp $
"""

import time,re,utctime

from w2lapp.schema.syntaxes import \
  LDAPSyntax,Boolean,DistinguishedName,DirectoryString,GeneralizedTime, \
  Integer,OctetString,SelectList,SecondsSinceEpoch,IA5String,\
  DynamicValueSelectList,syntax_registry

from w2lapp.core import utf2display

syntax_registry.registerAttrType(
  SecondsSinceEpoch.oid,[
    '1.3.6.1.4.1.7165.2.1.3', # pwdLastSet
    '1.3.6.1.4.1.7165.2.1.5', # logonTime
    '1.3.6.1.4.1.7165.2.1.6', # logoffTime
    '1.3.6.1.4.1.7165.2.1.7', # kickoffTime
    '1.3.6.1.4.1.7165.2.1.8', # pwdCanChange
    '1.3.6.1.4.1.7165.2.1.9', # pwdMustChange
    '1.3.6.1.4.1.7165.2.1.27', # sambaPwdLastSet
    '1.3.6.1.4.1.7165.2.1.28', # sambaPwdCanChange
    '1.3.6.1.4.1.7165.2.1.29', # sambaPwdMustChange
    '1.3.6.1.4.1.7165.2.1.30', # sambaLogonTime
    '1.3.6.1.4.1.7165.2.1.31', # sambaLogoffTime
    '1.3.6.1.4.1.7165.2.1.32', # sambaKickoffTime
  ]
)


from w2lapp.schema.plugins.activedirectory import LogonHours

syntax_registry.registerAttrType(
  LogonHours.oid,[
    '1.3.6.1.4.1.7165.2.1.55', # sambaLogonHours
  ]
)

class SambaAcctFlags(IA5String):
  oid = 'SambaAcctFlags-oid'
  desc = 'Samba 3 account flags'
  reObj=re.compile('^\[[NDHTUMWSLXI ]{11}\]$')
  flags_dict = {
    'N':'<b>N</b>o password.',
    'D':'<b>D</b>isabled.',
    'H':'<b>H</b>omedir required.',
    'T':'<b>T</b>emp account.', 
    'U':'<b>U</b>ser account (normal.', 
    'M':'<b>M</b>NS logon user account.', 
    'W':'<b>W</b>orkstation account.', 
    'S':'<b>S</b>erver account.', 
    'L':'<b>L</b>ocked account.', 
    'X':'No <b>X</b>piry on password', 
    'I':'<b>I</b>nterdomain trust account.',
  }

  def displayValue(self,valueindex=0,commandbutton=0):
    flags = self.attrValue[1:-1] # trim brackets
    table_rows = [
      '<tr><td>%s</td><td>%s</td></tr>\n' % ({1:'*',0:''}[f in flags],d)
      for f,d in self.flags_dict.items()
    ]
    return '<pre>%s</pre><table>\n%s\n</table>\n' % (
      utf2display(self._form.accept_charset,self.attrValue),
      ''.join(table_rows)
    )

syntax_registry.registerAttrType(
  SambaAcctFlags.oid,[
    '1.3.6.1.4.1.7165.2.1.26', # sambaAcctFlags
    '1.3.6.1.4.1.7165.2.1.4',  # acctFlags
  ]
)


class SambaSID(IA5String):
  oid = 'SambaSID-oid'
  desc = 'Samba 3 account flags'
  reObj=re.compile('^S(-[0-9]+)+$')

syntax_registry.registerAttrType(
  SambaSID.oid,[
    '1.3.6.1.4.1.7165.2.1.20', # sambaSID
    '1.3.6.1.4.1.7165.2.1.51', # sambaSIDList
  ]
)


class SambaForceLogoff(SelectList):
  oid = 'SambaForceLogoff-oid'
  desc = 'Disconnect Users outside logon hours (default: -1 => off, 0 => on)'
  attr_value_dict = {
    '':'',
    '0':'on',
    '-1':'off',
  }

syntax_registry.registerAttrType(
  SambaForceLogoff.oid,[
    '1.3.6.1.4.1.7165.2.1.66', # sambaForceLogoff
  ]
)


class SambaGroupType(SelectList):
  oid = 'SambaGroupType-oid'
  desc = 'Samba group type'
  attr_value_dict = {
    '':'',
    '2':'Domain Group',
    '4':'Local Group (Alias)',
    '5':'Built-in Group (well-known)',
  }

syntax_registry.registerAttrType(
  SambaGroupType.oid,[
    '1.3.6.1.4.1.7165.2.1.19', # sambaGroupType
  ]
)


class ReferencedSID(SambaSID):
  oid = 'ReferencedSID-oid'
  desc = 'SID which points to another object'

  def displayValue(self,valueindex=0,commandbutton=0):
    if commandbutton:
      search_anchor = self._form.applAnchor(
          'searchform','&raquo;',self._sid,
          [
            ('dn',self._dn),
            ('search_filterstr', '(sambaSid=%s)' % (
                utf2display(self._form.accept_charset,self.attrValue)
              )
            ),
            ('searchform_mode','exp'),
          ],
          title='Search by SID',
      )
    else:
      search_anchor = ''
    return '%s %s\n' % (
      SambaSID.displayValue(self,valueindex,commandbutton),
      search_anchor,
    )

syntax_registry.registerAttrType(
  ReferencedSID.oid,[
    '1.3.6.1.4.1.7165.2.1.23', # sambaPrimaryGroupSID
    '1.3.6.1.4.1.7165.2.1.51', # sambaSIDList
  ]
)


# Register all syntax classes in this module
for name in dir():
  syntax_registry.registerSyntaxClass(eval(name))
