"""
w2lapp.schema.syntaxes: classes for known attribute types

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: syntaxes.py,v 1.167 2009/08/07 14:11:57 michael Exp $
"""

import re,imghdr,sndhdr,urllib,time,utctime,ldap,ldapurl,pyweblib.forms, \
       ldaputil.base,ldap.schema, \
       w2lapp.viewer,w2lapp.form,w2lapp.gui,mspki.util,w2lapp.cnf

from types import StringType,UnicodeType,IntType,LongType,FloatType,ClassType
from sets import Set

from w2lapp.core import utf2display

from ldaputil.base import is_dn

# Detect Python Imaging Library (PIL)
try:
  from PIL import Image as PILImage
except ImportError:
  PILImage = None

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO

try:
  from ldapoidreg import oid as oid_desc_reg
except ImportError:
  oid_desc_reg = {}


class SyntaxRegistry:

  def __init__(self):
    self.oid2syntax = ldap.cidict.cidict()
    self.at2syntax = ldap.cidict.cidict()

  def registerSyntaxClass(self,c):
    if type(c) is ClassType and hasattr(c,'oid'):
      self.oid2syntax[c.oid] = c

  def registerAttrType(self,syntax_oid,attrTypes):
    for a in attrTypes:
      self.at2syntax[a.strip()] = syntax_oid

  def syntaxClass(self,schema,attrtype_nameoroid):
    if attrtype_nameoroid is None:
      return LDAPSyntax
    attrtype_oid = schema.getoid(ldap.schema.AttributeType,attrtype_nameoroid)
    if self.at2syntax.has_key(attrtype_oid):
      syntax_class = self.oid2syntax[self.at2syntax[attrtype_oid]]
    else:
      attrtype_se = schema.get_inheritedobj(ldap.schema.AttributeType,attrtype_oid,['syntax'])
      if attrtype_se is None or attrtype_se.syntax is None:
        syntax_class = LDAPSyntax
      else:
        syntax_class = self.oid2syntax.get(attrtype_se.syntax,LDAPSyntax)
    return syntax_class

  def attrInstance(self,sid,form,ls,dn,schema,attrType,attrValue,entry=None):
    syntax_class = self.syntaxClass(schema,attrType)
    attr_instance = syntax_class(sid,form,ls,dn,schema,attrType,attrValue,entry)
    return attr_instance
    

url_pattern  = r'^(ftp|http|https|news|snews|ldap|ldaps|mailto):(|//)[^ ]*'
url_regex  = re.compile(url_pattern)
labeleduri_regex = re.compile(url_pattern+r' .*')
timestamp_pattern = r'^([0-9]){12,14}(\.[0-9]+)*(Z|(\+|-)[0-9]{4})$'
timestamp_regex  = re.compile(timestamp_pattern)
mail_pattern = r'^[\w@.+=/_ -]+@[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$'
mail_regex = re.compile(mail_pattern)


####################################################################
# Classes of known syntaxes
####################################################################


class LDAPSyntaxValueError(ValueError):
  pass


class LDAPSyntaxRegexNoMatch(LDAPSyntaxValueError):
  pass


class LDAPSyntax:
  oid=''
  desc='Any LDAP syntax'
  maxLen = w2lapp.cnf.misc.input_maxfieldlen
  maxValues = w2lapp.cnf.misc.input_maxattrs
  mimeType = 'application/octet-stream'
  fileExt = 'bin'
  editable = 1
  charset = 'utf-8'
  reObj = None

  def __init__(self,sid,form,ls,dn,schema,attrType,attrValue,entry=None):
    assert entry is None or isinstance(entry,ldap.schema.models.Entry), \
      TypeError('entry must be ldap.schema.models.Entry but is %s' % (entry.__class__.__name__))
    self.attrType = attrType
    assert type(attrValue)==StringType or attrValue is None, "Argument 'attrValue' must be StringType or None"
    self.attrValue = attrValue
    self._sid = sid
    self._form = form
    self._ls = ls
    assert type(dn)==UnicodeType, "Argument 'dn' must be UnicodeType"
    self._dn = dn
    self._schema = schema
    self._entry = entry or {}

  def setAttrValue(self,attrValue):
    self.validate(attrValue)
    self.attrValue = attrValue

  def sanitizeInput(self,attrValue):
    return attrValue

  def _regexValidate(self,attrValue):
    if self.reObj and (self.reObj.match(attrValue) is None):
      raise LDAPSyntaxRegexNoMatch, \
        "Class %s: %s does not match pattern %s." % (
          self.__class__.__name__,repr(attrValue[0:]),repr(self.reObj.pattern)
        )
    return # _regexValidate()

  def _validate(self,attrValue):
    return 1

  def validate(self,attrValue):
    if attrValue:
      if not self._validate(attrValue):
        raise LDAPSyntaxValueError, \
          "Class %s: %s does not comply to syntax." % (
            self.__class__.__name__,repr(attrValue)
          )
      self._regexValidate(attrValue)

  def formValue(self):
    return unicode(self.attrValue or '',self.charset)

  def formField(self):
    input_field = pyweblib.forms.Input(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,None,default=None,size=min(self.maxLen,50)
    )
    input_field.charset = self._form.accept_charset
    input_field.setDefault(self.formValue())
    return input_field

  def getMimeType(self):
    return self.mimeType

  def displayValue(self,valueindex=0,commandbutton=0):
    if ldapurl.isLDAPUrl(self.attrValue):
      displayer_class = LDAPUrl
    elif url_regex.search(self.attrValue)!=None:
      displayer_class = Uri
    elif timestamp_regex.match(self.attrValue)!=None:
      displayer_class = GeneralizedTime
    elif mail_regex.match(self.attrValue)!=None:
      displayer_class = RFC822Address
    else:
      displayer_class = DirectoryString
    # Crude hack
    self_class = self.__class__
    self.__class__ = displayer_class
    result = displayer_class.displayValue(self,valueindex,commandbutton)
    self.__class__ = self_class
    return result


class Binary(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.5'
  desc = 'Binary'
  editable = 0

  def formField(self):
    f = pyweblib.forms.File(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,None,default=self.attrValue,size=50
    )
    f.mimeType = self.mimeType
    return f

  def displayValue(self,valueindex=0,commandbutton=0):
    return self._form.applAnchor(
      'read','View/Load',self._sid,
      [('dn',self._dn),('read_attr',self.attrType),('read_attrindex',str(valueindex))]
    )


class Audio(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.4'
  desc = 'Audio'
  mimeType = 'audio/basic'
  fileExt = 'au'

  def _validate(self,attrValue):
    f = StringIO(attrValue)
    res = sndhdr.test_au(attrValue,f)
    return res!=None

  def displayValue(self,valueindex=0,commandbutton=0):
    mimetype = self.getMimeType()
    return """
      <embed
        type="%s"
        autostart="false"
        src="%s/read/%s?dn=%s&amp;read_attr=%s&amp;read_attrindex=%d"
      >
      %d bytes of audio data (%s)
      """ % (
        mimetype,
        self._form.script_name,self._sid,
        urllib.quote(self._dn.encode(self._form.accept_charset)),
        urllib.quote(self.attrType),
        valueindex,
        len(self.attrValue),
        mimetype
      )


class Certificate(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.8'
  desc = 'X.509 Certificate'
  mimeType = 'application/pkix-cert'
  fileExt = 'cer'

  def sanitizeInput(self,attrValue):
    if mspki.util.pem_re.match(attrValue):
      return mspki.util.pem2der(attrValue)
    else:
      return attrValue

  def getMimeType(self):
    env = self._form.env
    if self._form.browser_type in ['Mozilla','Opera']:
      return 'application/x-x509-email-cert'
    return self.mimeType

  def displayValue(self,valueindex=0,commandbutton=0):
    return self._form.applAnchor(
      'read','View/Load',self._sid,
      [
        ('dn',self._dn),
        ('read_attr',self.attrType),
        ('read_attrindex',str(valueindex)),
        ('read_attrmode','view'),
      ]
    )


class CACertificate(Certificate):
  oid = 'CACertificate-oid'
  desc = 'X.509 CA Certificate'
  mimeType = 'application/x-x509-ca-cert'

  def getMimeType(self):
    return self.mimeType


class CertificateRevocationList(Certificate):
  oid = '1.3.6.1.4.1.1466.115.121.1.9'
  desc = 'Certificate Revocation List'
  mimeType = 'application/pkix-crl'
  fileExt = 'crl'

  def getMimeType(self):
    env = self._form.env
    browsertype,browserversion = pyweblib.helper.BrowserType(env.get('HTTP_USER_AGENT',''))
    if browsertype in ['Mozilla','Opera']:
      return 'application/x-pkcs7-crl'
    return self.mimeType


class DirectoryString(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.15'
  desc = 'Directory String'
  charset = 'utf-8'

  def _validate(self,attrValue):
    try:
      unicode(attrValue,self.charset).encode(self.charset)
    except UnicodeError:
      return 0
    else:
      return 1

  def sanitizeInput(self,attrValue):
    if type(attrValue)==StringType:
      return attrValue
    elif type(attrValue)==UnicodeType:
      return attrValue.encode(self.charset)
    else:
      raise ValueError, "Argument attrValue must be either StringType or UnicodeType"

  def displayValue(self,valueindex=0,commandbutton=0):
    return utf2display(self._form.accept_charset,self.attrValue)


class DistinguishedName(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.12'
  desc = 'Distinguished Name'

  def _validate(self,attrValue):
    return is_dn(attrValue)

  def displayValue(self,valueindex=0,commandbutton=0):
    return w2lapp.gui.DisplayDN(
      self._sid,
      self._form,
      self._ls,
      unicode(self.attrValue,self._ls.charset),
      commandbutton=commandbutton
    )


class NameAndOptionalUID(DistinguishedName):
  oid = '1.3.6.1.4.1.1466.115.121.1.34'
  desc = 'Name And Optional UID'

  def _splitDNandUID(self,v):
    try:
      sep_ind = v.rindex('#')
    except ValueError:
      dn = v
      uid = None
    else:
      dn = v[0:sep_ind]
      uid = v[sep_ind+1:]
    return dn,uid

  def _validate(self,attrValue):
    dn,uid = self._splitDNandUID(attrValue)
    return is_dn(dn)

  def displayValue(self,valueindex=0,commandbutton=0):
    value = self.attrValue.split('#')
    dn_str = w2lapp.gui.DisplayDN(
      self._sid,self._form,self._ls,unicode(value[0],self._ls.charset),commandbutton=commandbutton
    )
    if len(value)==1 or not value[1]:
      return dn_str
    else:
      return ' | '.join([value[1],dn_str])
      

class Boolean(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.7'
  desc = 'Boolean'

  def _validate(self,attrValue):
    v = attrValue.upper()
    return (v=='TRUE') or (v=='FALSE')

  def displayValue(self,valueindex=0,commandbutton=0):
    return self.attrValue.encode('ascii')

  def formField(self):
    if (self.attrValue or '').upper()==self.attrValue:
      options = ('','TRUE','FALSE')
    else:
      options = ('','true','false')
    return pyweblib.forms.Select(
      self.attrType,
      ': '.join([self.attrType,self.desc]),1,
      options=options,required=0,default=self.attrValue
    )


class CaseExactString(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.26'
  desc = 'Case Exact String'


class BitString(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.6'
  desc = 'Bit String'
  reObj=re.compile("^'[01]+'B$")


class GeneralizedTime(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.24'
  desc = 'Generalized Time'
  reObj=timestamp_regex

  def displayValue(self,valueindex=0,commandbutton=0):
    try:
      return utctime.strftimeiso8601(utctime.strptime(self.attrValue)).encode('ascii')
    except ValueError:
      return utf2display(self._form.accept_charset,self.attrValue)


class UTCTime(GeneralizedTime):
  oid = '1.3.6.1.4.1.1466.115.121.1.53'
  desc = 'UTC Time'


class IA5String(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.26'
  desc = 'IA5 String'
  charset = 'ascii'


class NullTerminatedDirectoryString(DirectoryString):
  oid = 'NullTerminatedDirectoryString-oid'
  desc = 'Directory String terminated by null-byte'

  def sanitizeInput(self,attrValue):
    return attrValue+chr(0)

  def _validate(self,attrValue):
    return attrValue.endswith(chr(0))

  def formValue(self):
    return unicode((self.attrValue or chr(0))[:-1],self.charset)

  def displayValue(self,valueindex=0,commandbutton=0):
    return utf2display(self._form.accept_charset,(self.attrValue or chr(0))[:-1])


class OtherMailbox(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.39'
  desc = 'Other Mailbox'
  charset = 'ascii'


class Integer(IA5String):
  oid = '1.3.6.1.4.1.1466.115.121.1.27'
  desc = 'Integer'

  def sanitizeInput(self,attrValue):
    try:
      return str(int(attrValue))
    except ValueError:
      return attrValue

  def _validate(self,attrValue):
    try:
      int(attrValue)
    except ValueError:
      return 0
    else:
      return 1


class IntegerRange(Integer):
  oid = 'IntegerRange-oid'
  desc = 'Integer within a certain range'
  # The following class members are set to None to make direct use of this class fail
  minValue = None
  maxValue = None

  def _validate(self,attrValue):
    try:
      intValue = int(attrValue)
    except ValueError:
      return 0
    else:
      return (intValue>=self.minValue) and (intValue<=self.maxValue)

  def formField(self):
    max_len = max(len(str(self.minValue)),len(str(self.maxValue)))    
    return pyweblib.forms.Input(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      max_len,self.maxValues,'[0-9]*',default=self.attrValue,size=max_len
    )


class RFC822Address(IA5String):
  oid = 'RFC822Address-oid'
  desc = 'RFC 822 mail address'
  reObj= re.compile(mail_pattern)

  def displayValue(self,valueindex=0,commandbutton=0):
    return w2lapp.gui.MailToURL(self.attrValue)


class Uri(DirectoryString):
  oid = 'Uri-OID'
  desc = 'URI'
  reObj = url_regex

  def displayValue(self,valueindex=0,commandbutton=0):
    try:
      url,label = self.attrValue.split(' ',1)
    except ValueError:
      url,label = self.attrValue,self.attrValue
      display_url = ''
    else:
      display_url = ' (%s)' % (url)
    return '<a href="%s/urlredirect/%s?%s">%s%s</a>' % (
      self._form.script_name,
      self._sid,
      url.encode('ascii'),
      utf2display(self._form.accept_charset,label),
      utf2display(self._form.accept_charset,display_url),
    )


class Image(Binary):
  oid = 'Image-OID'
  desc = 'Image base class'
  mimeType = 'application/octet-stream'
  fileExt = 'bin'
  imageFormat = None

  def _validate(self,attrValue):
    return imghdr.test_jpeg(attrValue,None)==self.imageFormat.lower()

  def sanitizeInput(self,attrValue):
    if not self._validate(attrValue) and PILImage:
      f = StringIO(attrValue)
      f2 = StringIO()
      try:
        try:
          im = PILImage.open(f)
          im.save(f2,self.imageFormat)
        except (IOError,ValueError),e:
          attrValue = None
        else:
          attrValue = f2.getvalue()
      finally:
        f.close()
    return attrValue

  def displayValue(self,valueindex=0,commandbutton=0):
    maxwidth,maxheight = 100,150
    width,height = None,None
    size_attr_html = ''
    if PILImage:
      f = StringIO(self.attrValue)
      try:
        im = PILImage.open(f)
      except IOError:
        pass
      else:
        width,height = im.size
        if width>maxwidth:
          size_attr_html = 'width="%d" height="%d"' % (maxwidth,int(float(maxwidth)/width*height))
        elif height>maxheight:
          size_attr_html = 'width="%d" height="%d"' % (int(float(maxheight)/height*width),maxheight)
        else:
          size_attr_html = 'width="%d" height="%d"' % (width,height)
    attr_value_len = len(self.attrValue)
    img_link = "%s/read/%s?dn=%s&amp;read_attr=%s&amp;read_attrindex=%d&amp;read_attrmode=load&amp;read_attrmimetype=image/jpeg" % (
      self._form.script_name,self._sid,
      urllib.quote(self._dn.encode(self._form.accept_charset)),
      urllib.quote(self.attrType),
      valueindex,
    )
    if attr_value_len<=0:
      return """
        <a href="%s">
          <img
            src="data:%s;base64,%s"
            alt="%d bytes of image data">
        </a>
        """ % (img_link,self.mimeType,self.attrValue.encode('base64'),attr_value_len)
    else:
      return """
        <a href="%s">
          <img
            src="%s"
            alt="%d bytes of image data" %s>
        </a>
        """ % (
        img_link,img_link,attr_value_len,size_attr_html,
      )


class JPEGImage(Image):
  oid = '1.3.6.1.4.1.1466.115.121.1.28'
  desc = 'JPEG image'
  mimeType = 'image/jpeg'
  fileExt = 'jpg'
  imageFormat = 'JPEG'


class PhotoG3Fax(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.23'
  desc = 'Photo (G3 fax)'
  mimeType = 'image/g3fax'
  fileExt = 'tif'


class OID(IA5String):
  oid = '1.3.6.1.4.1.1466.115.121.1.38'
  desc = 'OID'
  regex = '([0-9]+.)*[0-9]'

  def displayValue(self,valueindex=0,commandbutton=0):
    attrValue = self.attrValue.encode('ascii')
    try:
      name,description,reference = oid_desc_reg[attrValue]
    except (KeyError,ValueError):
      return attrValue
    else:
      return '<strong>%s</strong> (%s):<br>%s (see %s)' % (
        utf2display(self._form.accept_charset,name),
        IA5String.displayValue(self,valueindex,commandbutton),
        utf2display(self._form.accept_charset,description),
        utf2display(self._form.accept_charset,reference)
      )


class LDAPUrl(DirectoryString):
  oid = 'LDAPUrl-oid'
  desc = 'LDAP URL'

  def displayValue(self,valueindex=0,commandbutton=0):
    try:
      if commandbutton:
        commandbuttonstr = w2lapp.gui.LDAPURLButton(self._sid,self._form,self._ls,self.attrValue)
      else:
        commandbuttonstr = ''
      return '<table><tr><td>%s</td><td><a href="%s">%s</a></td></tr></table>' % (
               commandbuttonstr,
               self.attrValue.encode(self._form.accept_charset),
               utf2display(self._form.accept_charset,self.attrValue)
             )
    except ValueError:
      return '<strong>Not a valid LDAP URL:</strong> %s' % utf2display(self._form.accept_charset,repr(self.attrValue))


class OctetString(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.40'
  desc = 'Octet String'
  editable = 1
  minInputRows = 1  # minimum number of rows for input field
  maxInputRows = 15 # maximum number of rows for in input field

  def sanitizeInput(self,attrValue):
    for c in [':',' ',',','\r','\n']:
      attrValue = attrValue.replace(c,'')
    result = []
    try:
      for i in range(0,len(attrValue),2):
        result.append(chr(int(attrValue[i:i+2],16)))
    except ValueError:
      raise LDAPSyntaxValueError, 'Illegal character in human-readable OctetString representation'
    else:
      result_str = ''.join(result)
    return result_str

  def displayValue(self,valueindex=0,commandbutton=0):
    return '<code><font size="-1">%s</font></code>' % mspki.util.HexString(
      self.attrValue,
      delimiter=':',wrap=64,linesep='<br>'
    )

  def formValue(self):
    return mspki.util.HexString(str(self.attrValue or ''),delimiter=':',wrap=48,linesep='\r\n')

  def formField(self):
    form_value = self.formValue()
    return pyweblib.forms.Textarea(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      10000,1,
      None,
      default=form_value,
      rows=max(self.minInputRows,min(self.maxInputRows,form_value.count('\r\n'))),
      cols=49
    )


class MultilineText(DirectoryString):
  oid = 'MultilineText-oid'
  desc = 'Multiple lines of text'
  reObj=re.compile('^.*$',re.S+re.M)
  lineSep = '\r\n'
  mimeType = 'text/plain'
  cols = 66
  minInputRows = 1  # minimum number of rows for input field
  maxInputRows = 30 # maximum number of rows for in input field

  def _split_lines(self,v):
    return v.split(self.lineSep)

  def sanitizeInput(self,inputValue):
    return inputValue.replace('\r','').replace('\n',self.lineSep)

  def displayValue(self,valueindex=0,commandbutton=0):
    lines = [
      utf2display(self._form.accept_charset,l)
      for l in self._split_lines(self.attrValue)
    ]
    return '<br>'.join(lines)

  def formValue(self):
    splitted_lines = self._split_lines(unicode(self.attrValue or '',self._ls.charset))
    return u'\r\n'.join(splitted_lines)

  def formField(self):
    form_value=self.formValue()
    return pyweblib.forms.Textarea(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,
      None,
      default=form_value,
      rows=max(self.minInputRows,min(self.maxInputRows,form_value.count('\r\n'))),
      cols=self.cols
    )


class PostalAddress(MultilineText):
  oid = '1.3.6.1.4.1.1466.115.121.1.41'
  desc = 'Postal Address'
  lineSep = ' $ '
  cols = 40

  def _split_lines(self,value):
    return [ v.strip() for v in value.split(self.lineSep.strip()) ]

  def sanitizeInput(self,inputValue):
    return inputValue.replace('\r','').replace('\n',self.lineSep)


class PrintableString(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.44'
  desc = 'Printable String'
  reObj= re.compile('^[a-zA-Z0-9"()+,.=/:? -]*$')
  charset = 'ascii'


class NumericString(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.36'
  desc = 'Numeric String'
  reObj= re.compile('^[ 0-9]+$')


class EnhancedGuide(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.21'
  desc = 'Enhanced Search Guide'


class Guide(EnhancedGuide):
  oid = '1.3.6.1.4.1.1466.115.121.1.25'
  desc = 'Search Guide'


class TelephoneNumber(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.50'
  desc = 'Telephone Number'
  reObj= re.compile('^[0-9+x(). /-]+$')


class FacsimileTelephoneNumber(TelephoneNumber):
  oid = '1.3.6.1.4.1.1466.115.121.1.22'
  desc = 'Facsimile Number'
  reObj= re.compile('^[0-9+x(). /-]+(\$(twoDimensional|fineResolution|unlimitedLength|b4Length|a3Width|b4Width|uncompressed))*$')


class TelexNumber(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.52'
  desc = 'Telex Number'


class TeletexTerminalIdentifier(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.51'
  desc = 'Teletex Terminal Identifier'


class ObjectGUID(LDAPSyntax):
  oid = 'ObjectGUID-oid'
  desc = 'Object GUID'
  charset = 'ascii'

  def displayValue(self,valueindex=0,commandbutton=0):
    objectguid_str = ''.join(['%02X' % ord(c) for c in self.attrValue])
    return ldapurl.LDAPUrl(
      ldapUrl=self._ls.uri,
      dn='GUID=%s' % (objectguid_str),
      who=None,cred=None
    ).htmlHREF(
      hrefText=objectguid_str,
      hrefTarget=None
    )


class SecondsSinceEpoch(Integer):
  oid = 'SecondsSinceEpoch-oid'
  desc = 'Seconds since epoch (1970-01-01 00:00:00)'

  def displayValue(self,valueindex=0,commandbutton=0):
    try:
      return utctime.strftimeiso8601(time.gmtime(float(self.attrValue))).encode('ascii')
    except ValueError:
      return utf2display(self._form.accept_charset,self.attrValue)


class DaysSinceEpoch(Integer):
  oid = 'DaysSinceEpoch-oid'
  desc = 'Days since epoch (1970-01-01)'

  def displayValue(self,valueindex=0,commandbutton=0):
    try:
      return utctime.strftimeiso8601(time.gmtime(float(self.attrValue)*86400)).encode('ascii')
    except ValueError:
      return utf2display(self._form.accept_charset,self.attrValue)


class SelectList(LDAPSyntax,DirectoryString):
  """
  Base class for dictionary based select lists which
  should not be used directly
  """
  attr_value_dict = {} # Mapping attribute value to attribute description
  
  def _sortedSelectOptions(self):
    current_attr_values = Set(self._entry.get(self.attrType,[]))
    tmplist = [
      (x[1], x)
      for x in self.attr_value_dict.items()
      if x[0]==self.attrValue or not x[0] in current_attr_values
    ]
    tmplist.sort()
    return [x for (k,x) in tmplist ]
    
  def _validate(self,attrValue):
    return self.attr_value_dict.has_key(attrValue)

  def displayValue(self,valueindex=0,commandbutton=0):
    attr_value_str = utf2display(self._form.accept_charset,self.attrValue)
    try:
      attr_value_desc=self.attr_value_dict[self.attrValue]
    except KeyError:
      return attr_value_str
    else:
      if attr_value_desc==self.attrValue:
        return attr_value_str
      else:
        return ': '.join((
          utf2display(self._form.accept_charset,attr_value_desc),attr_value_str
        ))

  def formField(self):
    if not self.attr_value_dict or self.attr_value_dict=={'':''}:
      return DirectoryString.formField(self)
    else:
      f = pyweblib.forms.Select(
        self.attrType,
        ': '.join([self.attrType,self.desc]),1,
        options=self._sortedSelectOptions(),required=0,default=self.attrValue
      )
      f.charset=self._form.accept_charset
      return f


class DynamicValueSelectList(SelectList):
  ldap_url = None

  def __init__(self,sid,form,ls,dn,schema,attrType,attrValue,entry=None):
    self.lu_obj = ldapurl.LDAPUrl(self.ldap_url)
    SelectList.__init__(self,sid,form,ls,dn,schema,attrType,attrValue,entry)
    self._doSearch()

  def _determineSearchDN(self,current_dn,ldap_url_dn):
    if ldap_url_dn=='_':
      result_dn = self._ls.getSearchRoot(self._dn)
    elif ldap_url_dn=='.':
      result_dn = current_dn
    elif ldap_url_dn=='..':
      result_dn = ldaputil.base.ParentDN(current_dn)
    elif ldap_url_dn.endswith(',_'):
      result_dn = ','.join((ldap_url_dn[:-2],self._ls.getSearchRoot(self._dn)))
    elif ldap_url_dn.endswith(',.'):
      result_dn = ','.join((ldap_url_dn[:-2],current_dn))
    elif ldap_url_dn.endswith(',..'):
      result_dn = ','.join((ldap_url_dn[:-3],ldaputil.base.ParentDN(current_dn)))
    else:
      result_dn = ldap_url_dn
    return result_dn # _determineSearchDN()

  def _doSearch(self,initial_attr_value_dict=None):
    # Enable empty value in any case
    self.attr_value_dict = initial_attr_value_dict or {'':'- not set -'}
    if self.lu_obj.hostport:
      # New connection to separate server
      # not implemented yet!
      pass
    else:
      search_dn = self._determineSearchDN(self._dn,self.lu_obj.dn)
      search_scope = self.lu_obj.scope or ldap.SCOPE_BASE
      # Use the existing LDAP connection as current user
      try:
        ldap_result = self._ls.l.search_s(
          search_dn,
          search_scope,
          self.lu_obj.filterstr or '(objectClass=*)',
          self.lu_obj.attrs or ['objectClass']
        )
      except ldap.NO_SUCH_OBJECT:
        return
    if search_scope==ldap.SCOPE_BASE:
      dn_r,entry_r=ldap_result[0]
      # When reading a single entry we build the map from a single multi-valued attribute
      assert len(self.lu_obj.attrs or [])==1,"attrlist in ldap_url must be of length 1 if scope is base"
      list_attr = self.lu_obj.attrs[0]
      attr_values_u = [
        unicode(attr_value,self._ls.charset)
        for attr_value in entry_r[list_attr]
      ]
      self.attr_value_dict=dict([ (u,u) for u in attr_values_u ])
    else:
      try:
        option_value_map,option_text_map = (self.lu_obj.attrs or [])[:2]
      except ValueError:
        option_value_map,option_text_map = (None,(self.lu_obj.attrs or [None])[0])
      for dn_r,entry_r in ldap_result:
        # Check whether it's a real search result (ignore search continuations)
        if not dn_r is None:
          entry_r[None] = [dn_r]
          try:
            self.attr_value_dict[unicode(entry_r[option_value_map][0],self._ls.charset)] = unicode(entry_r.get(option_text_map,entry_r[option_value_map])[0],self._ls.charset)
          except KeyError:
            pass
    return # _doSearch()


class DynamicDNSelectList(DynamicValueSelectList,DistinguishedName):

  def displayValue(self,valueindex=0,commandbutton=0):
    try:
      attr_value_desc=self.attr_value_dict[self.attrValue]
    except KeyError:
      display_text = ''
    else:
      if attr_value_desc==self.attrValue:
        display_text = ''
      else:
        display_text = utf2display(self._form.accept_charset,attr_value_desc+': ')
    return ''.join((
      display_text,
      DistinguishedName.displayValue(self,valueindex,commandbutton)
    ))


class SchemaElementSelectList(SelectList):
  schema_class = None

  def __init__(self,sid,form,ls,dn,schema,attrType,attrValue,entry=None):
    self.attr_value_dict = {'':''}
    SelectList.__init__(self,sid,form,ls,dn,schema,attrType,attrValue)


class CountryString(SelectList):
  oid = '1.3.6.1.4.1.1466.115.121.1.11'
  desc = 'Two letter country string as listed in ISO 3166-2'
  attr_value_dict = ldap.cidict.cidict(w2lapp.cnf.countries.c_dict)


class DeliveryMethod(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.14'
  desc = 'Delivery Method'
  pdm = '(any|mhs|physical|telex|teletex|g3fax|g4fax|ia5|videotex|telephone)'
  reObj= re.compile('^%s[ $]*%s$' % (pdm,pdm))


class BitArrayInteger(MultilineText,Integer):
  oid = 'BitArrayInteger-oid'
  flag_desc_table = ()
  true_false_desc={1:'+',0:'-'}
  minValue=0

  def __init__(self,sid,form,ls,dn,schema,attrType,attrValue,entry=None):
    Integer.__init__(self,sid,form,ls,dn,schema,attrType,attrValue)
    self.flag_desc2int=dict(self.flag_desc_table)
    self.flag_int2desc=dict([(j,i) for i,j in self.flag_desc_table])
    self.maxValue=sum([j for i,j in self.flag_desc_table])
    self.minInputRows=maxInputRows=max(len(self.flag_desc_table),1)

  def sanitizeInput(self,inputValue):
    result = 0
    for row in inputValue.split('\n'):
      row=row.strip()
      try:
        flag_set,flag_desc=row[0],row[1:]
      except IndexError:
        pass
      else:
        if flag_set=='+':
          try:
            result=result|self.flag_desc2int[flag_desc]
          except KeyError:
            pass
    return str(result)

  def formValue(self):
    attr_value_int=int(self.attrValue or 0)
    flag_lines = [
      ''.join((
        self.true_false_desc[int((attr_value_int&flag_int)>0)],
        flag_desc
      ))
      for flag_desc,flag_int in self.flag_desc_table
    ]
    return u'\r\n'.join(flag_lines)

  def formField(self):
    form_value=self.formValue()
    return pyweblib.forms.Textarea(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,
      None,
      default=form_value,
      rows=max(self.minInputRows,min(self.maxInputRows,form_value.count('\n'))),
      cols=self.cols
    )

  def displayValue(self,valueindex=0,commandbutton=0):
    attrValue_int = int(self.attrValue)
    return """%s<br>
    <table summary="Flags">
    <tr><th>Property flag</th><th>Value</th><th>Status</th></tr>
    %s
    </table>
    """ % (
      Integer.displayValue(self,valueindex,commandbutton),
      '\n'.join([
        '<tr><td>%s</td><td>%s</td><td>%s</td></tr>' % (
          desc,
          hex(flag_value),
          {0:'-',1:'on'}[int((attrValue_int & flag_value)>0)]
        )
        for desc,flag_value in self.flag_desc_table
      ])
    )
  

class UUID(IA5String):
  oid = '1.3.6.1.1.16.1'
  desc = 'UUID'
  reObj = re.compile('^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$')


class DomainComponent(IA5String):
  oid = 'DomainComponent-oid'
  desc = 'DNS domain name component'
  reObj = re.compile('^[a-zA-Z0-9-]+$')


class DNSDomain(IA5String):
  oid = 'DNSDomain-oid'
  desc = 'DNS domain name (see RFC 1035)'
  reObj = re.compile('^[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$')


class YesNoIntegerFlag(SelectList):
  oid = 'YesNoIntegerFlag-oid'
  desc = '0 means no, 1 means yes'
  attr_value_dict = {
    '':'- not set -',
    '0':'no',
    '1':'yes',
  }


class OnOffFlag(DirectoryString):
  oid = 'OnOffFlag-oid'
  desc = 'Only values on or off are allowed'

  def _validate(self,attrValue):
    v = attrValue.lower()
    return (v=='on') or (v=='off')

  def displayValue(self,valueindex=0,commandbutton=0):
    return self.attrValue.encode('ascii')

  def formField(self):
    return pyweblib.forms.Select(
      self.attrType,
      ': '.join([self.attrType,self.desc]),1,
      options=('','on','off'),required=0,default=self.attrValue.lower()
    )


try:
  import xml.etree.ElementTree
except ImportError:
  pass
else:

  class XmlValue(MultilineText):
    oid = 'XmlValue-oid'
    desc = 'XML data'
    lineSep = '\n'
    mimeType = 'text/xml'
    cols = 64

    def _validate(self,attrValue):
      try:
        import xml.etree.ElementTree
      except ImportError:
        return 1
      else:
        try:
          xml.etree.ElementTree.XML(attrValue)
        except xml.parsers.expat.ExpatError:
          return 0
        else:
          return 1
          
    def sanitizeInput(self,attrValue):
      return attrValue

    def displayValue(self,valueindex=0,commandbutton=0):
      lines = [
        utf2display(self._form.accept_charset,l,tab_identiation='&nbsp;&nbsp;&nbsp;&nbsp;')
        for l in self._split_lines(self.attrValue)
      ]
      return '<code>%s</code>' % '<br>'.join(lines)


try:
  # Try to import optional module pisces
  from pisces import asn1
except ImportError:
  # Well, silently ignore and don't do anything
  pass
else:

  class ASN1Object(Binary):
    oid = 'ASN1Object-oid'
    desc = 'BER encoded ASN.1 data'

    def displayValue(self,valueindex=0,commandbutton=0):
      asn1obj = asn1.parse(self.attrValue)
      return ''.join((
        '<code>',
        str(asn1obj).replace('{','<br>{').replace('}','}<br>').replace('  ','&nbsp;&nbsp;'),
        '</code>'
      ))

  class DumpASN1CfgOID(OID):
    oid = 'DumpASN1Cfg-oid'
    desc = "OID registered in Peter Gutmann's dumpasn1.cfg"

    def displayValue(self,valueindex=0,commandbutton=0):
      attrValue = self.attrValue.encode('ascii')
      try:
        import pisces.asn1
        pisces_oid = pisces.asn1.OID(tuple(map(int,attrValue.split('.'))))
        desc = mspki.asn1helper.GetOIDDescription(
          pisces_oid,
          w2lapp.viewer.oids,
          includeoid=1
        )
      except ValueError:
        return self.attrValue
      else:
        return desc


# Set up the central syntax registry instance
syntax_registry = SyntaxRegistry()
# Register all syntax classes in this module
for symbol_name in dir():
  syntax_registry.registerSyntaxClass(eval(symbol_name))

