"""
w2lapp.search.py: do a search and return results in several formats

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: search.py,v 1.168 2009/08/06 13:35:46 michael Exp $
"""

import time,urllib,ldap,ldap.async,dsml, \
       msbase,pyweblib.forms,pyweblib.httphelper,ldaputil.base, \
       w2lapp.core,w2lapp.cnf,w2lapp.gui,w2lapp.searchform

from ldapurl import LDAPUrl

from msbase import GrabKeys
from sets import Set

SizeLimitMsg = """
<p>
  <strong>
    Only partial results received. Try to refine search.
  </strong><br>
  %s
</p>
"""

is_search_result = ldap.async._entryResultTypes

is_search_reference = {
  ldap.RES_SEARCH_REFERENCE:None
}

class DSMLWriter(ldap.async.AsyncSearchHandler):
  """
  Class for writing a stream LDAP search results to a DSML file
  """
  _entryResultTypes = ldap.async._entryResultTypes

  def __init__(self,l,f,base64_attrs=[],dsml_comment=''):
    ldap.async.AsyncSearchHandler.__init__(self,l)
    self._dsmlWriter = dsml.DSMLWriter(
      f,base64_attrs,dsml_comment
    )
    self._base64_attrs = base64_attrs
    self._indent_offset = 2

  def preProcessing(self):
    self._dsmlWriter.writeHeader()

  def postProcessing(self):
    self._dsmlWriter.writeFooter()

  def _processSingleResult(self,resultType,resultItem):
    if self._entryResultTypes.has_key(resultType):
      dn,entry = resultItem
      self._dsmlWriter.writeRecord(dn,entry)


def w2l_Search(
  sid,outf,command,form,ls,dn,
  scope,
  search_filterstr
):
  """
  Search for entries and output results as table, pretty-printable output
  or LDIF formatted
  """

  # This should speed up things
  utf2display = w2lapp.core.utf2display

  # Hmm, this retrieves sub schema sub entry for the search root.
  # Theoretically it could be different for all search results.
  # But what the hey...
  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  search_output = form.getInputValue('search_output',['table'])[0]
  search_opattrs = form.getInputValue('search_opattrs',['no'])[0]=='yes'

  scope = {None:ldap.SCOPE_SUBTREE}.get(scope,scope)

  if ('search_option' in form.inputFieldNames) or \
     ('search_attr' in form.inputFieldNames) or \
     ('search_string' in form.inputFieldNames):
    search_mode = form.getInputValue('search_mode',[r'(&%s)'])[0]
    search_option = form.getInputValue('search_option',[])
    search_attr = form.getInputValue('search_attr',[])
    search_string = form.getInputValue('search_string',[])
    if (len(search_option)==len(search_attr)==len(search_string)):
      # Get search mode
      # Build LDAP search filter from input data of advanced search form
      search_filter = []
      for i in range(len(form.field['search_option'].value)):
        search_string,search_option,search_attr = \
          form.field['search_string'].value[i], \
          form.field['search_option'].value[i], \
          form.field['search_attr'].value[i]
        if search_option==r'(%s=*)':
          search_filter.append(search_option % (search_attr))
        elif search_string:
          search_filter.append(search_option % (
            search_attr,ldap.filter.escape_filter_chars(search_string)
          ))
      # Teilsuchen werden alle und-verknuepft
      if len(search_filter)==1:
        search_filterstr = ''.join(search_filter)
      elif len(search_filter)>1:
        search_filterstr = search_mode % (''.join(search_filter))
      else:
        raise w2lapp.core.ErrorExit('Empty search form data.')
    else:
      raise w2lapp.core.ErrorExit('Invalid search form data.')
  else:
    search_filterstr = form.getInputValue(
      'search_filterstr',[search_filterstr or '(objectClass=*)']
    )[0]

  search_resminindex = int(
    form.getInputValue('search_resminindex',['0'])[0]
  )
  search_resnumber = int(
    form.getInputValue(
      'search_resnumber',
      [str(w2lapp.cnf.GetParam(ls,'search_resultsperpage',10))]
    )[0]
  )

  requested_attrs = w2lapp.cnf.GetParam(ls,'requested_attrs',[])

  if search_output=='print':
    print_template_filenames_dict = w2lapp.cnf.GetParam(ls,'print_template',None)
    print_cols = w2lapp.cnf.GetParam(ls,'print_cols','4')
    
    if print_template_filenames_dict is None:
      raise w2lapp.core.ErrorExit('No templates for printing defined.')

    read_attrs = ['objectclass']
    print_template_str_dict = msbase.CaseinsensitiveStringKeyDict()

    for oc in print_template_filenames_dict.keys():
      try:
        print_template_str_dict[oc] = open(print_template_filenames_dict[oc],'r').read()
      except IOError:
        pass
      else:
        read_attrs = Set(read_attrs)
        read_attrs.update(GrabKeys(print_template_str_dict[oc]).keys)
    result_handler = ldap.async.List(ls.l)

  elif search_output in ['table','raw']:

    search_tdtemplate = ldap.cidict.cidict(
      w2lapp.cnf.GetParam(ls,'search_tdtemplate',{})
    )
    search_tdtemplate_keys = search_tdtemplate.keys()
    search_tdtemplate_keys_lower = search_tdtemplate.data.keys()
    search_tablistattrs = w2lapp.cnf.GetParam(ls,'search_tablistattrs',[])

    search_attrs = form.getInputValue('search_attrs',[''])[0]
    if search_attrs:
      read_attrs = [ a.strip() for a in search_attrs.split(',') ]
    else:
      # Build the list of attributes to read from templates
      if not search_output=='raw':
        read_attrs = Set(['objectClass'])
        for oc in search_tdtemplate_keys:
          read_attrs.update(GrabKeys(search_tdtemplate[oc]).keys)
        if search_tablistattrs:
          read_attrs.update(search_tablistattrs)
      else:
        read_attrs = Set([])
      read_attrs.update([
        'subschemaSubentry','displayName','description','structuralObjectClass',
        'hasSubordinates','subordinateCount',
        'numSubordinates',
        'numAllSubordinates', #  Siemens DirX
        'countImmSubordinates','countTotSubordinates', # Critical Path Directory Server
        'msDS-Approx-Immed-Subordinates' # MS Active Directory
      ])
    # Create async search handler instance
    result_handler = ldap.async.List(ls.l)

  elif search_output in ['ldif','ldif1']:
    # read all attributes
    read_attrs = ({0:['*'],1:['*','+']}[ls.supportsAllOpAttr and search_opattrs]+requested_attrs) or None
    result_handler = ldap.async.LDIFWriter(ls.l,outf)

  elif search_output=='dsml1':
    # read all attributes
    read_attrs = ({0:['*'],1:['*','+']}[ls.supportsAllOpAttr and search_opattrs]+requested_attrs) or None
    result_handler = DSMLWriter(ls.l,outf)

  search_ldap_url = LDAPUrl(ls.uri)
  search_ldap_url.dn = dn.encode(ls.charset)
  search_ldap_url.who=(ls.who or u'').encode(ls.charset) or None
  search_ldap_url.scope = scope
  search_ldap_url.filterstr = search_filterstr.encode(ls.charset)

  if search_resnumber:
    search_size_limit = search_resminindex+search_resnumber
  else:
    search_size_limit = -1

  try:
    # Start the search
    result_handler.startSearch(
      dn.encode(ls.charset),
      scope,
      search_filterstr.encode(ls.charset),
      attrList=ldaputil.base.encode_unicode_list(read_attrs or []) or None,
      attrsOnly=0,
      sizelimit=search_size_limit
    )
  except ldap.FILTER_ERROR:
    # Give the user a chance to edit his bad search filter
    w2lapp.searchform.w2l_SearchForm(
      sid,outf,command,form,ls,dn,searchform_mode='exp',
      Msg='<p>Error: Bad search filter!</p>',
      search_filterstr=search_filterstr,
      scope=scope
    )
    return
  except ldap.NO_SUCH_OBJECT,e:
    if dn:
      raise e

  if search_output in ['table','raw']:

    SearchWarningMsg = ''
    partial_results = 0

    try:
      result_handler.processResults(
        search_resminindex,search_resnumber,timeout=ls.timeout
      )
    except (ldap.SIZELIMIT_EXCEEDED,ldap.ADMINLIMIT_EXCEEDED),e:
      if search_size_limit<0 or result_handler.endResultBreak<search_size_limit:
        SearchWarningMsg = SizeLimitMsg % (w2lapp.gui.LDAPError2ErrMsg(e))
      partial_results = 1
    except ldap.NO_SUCH_OBJECT,e:
      partial_results = 0
      if dn:
        raise e

    search_resminindex = result_handler.beginResultsDropped
    resind = result_handler.endResultBreak
    result_dnlist = result_handler.allResults

    # HACK! Searching the root level the namingContexts is
    # appended if not already received in search result
    if not dn and scope==ldap.SCOPE_ONELEVEL:
      d = ldap.cidict.cidict()
      for result_dn in ls.getNamingContexts():
        if result_dn:
          d[result_dn] = result_dn
      for r in result_dnlist:
        result_dn = r[1][0]
        if result_dn!=None and d.has_key(result_dn):
          del d[result_dn]
      result_dnlist.extend([
        (ldap.RES_SEARCH_ENTRY,(result_dn.encode(ls.charset),{}))
        for result_dn in d.values()
      ])

    result_dnlist.sort()

    ContextMenuList = []

    if result_dnlist:

      # There's still data left to be read
      if search_size_limit>0 or search_resminindex:

        prev_resminindex = max(0,search_resminindex-search_resnumber)

        if search_resminindex+search_resnumber<=resind:
          partial_results = 1
          ContextMenuList.append(
            form.applAnchor(
              'search','-&gt;&gt;',sid,
              [
                ('dn',dn),
                ('search_filterstr',search_filterstr),
                ('search_output',search_output),
                ('search_resminindex',str(search_resminindex+search_resnumber)),
                ('search_resnumber',str(search_resnumber)),
                ('scope',str(scope)),
              ],
              title='Display next search results %d to %d' % (1+search_resminindex+search_resnumber,search_resminindex+2*search_resnumber),
            )
          )

        if search_resminindex:
          partial_results = 1
          ContextMenuList.append(
            form.applAnchor(
              'search','&lt;&lt;-',sid,
              [
                ('dn',dn),
                ('search_filterstr',search_filterstr),
                ('search_output',search_output),
                ('search_resminindex',str(prev_resminindex)),
                ('search_resnumber',str(search_resnumber)),
                ('scope',str(scope)),
              ],
              title='Display previous search results %d to %d' % (1+prev_resminindex,prev_resminindex+search_resnumber),
            )
          )

        if partial_results:
          ContextMenuList.append(
            form.applAnchor(
              'search','Display all',sid,
              [
                ('dn',dn),
                ('search_filterstr',search_filterstr),
                ('search_output',search_output),
                ('search_resnumber','0'),
                ('scope',str(scope)),
              ],
              title='Display all search results at once',
            )
          )

        result_message = 'Results %d - %d of %s search with filter &quot;%s&quot;.' % (
            search_resminindex+1,
            resind,
            ldaputil.base.SEARCH_SCOPE_STR[scope],
            utf2display(form.accept_charset,search_filterstr)
        )

      else:
        result_message = '%s search with filter &quot;%s&quot; found %d entries.' % (
          ldaputil.base.SEARCH_SCOPE_STR[scope],
          utf2display(form.accept_charset,search_filterstr),
          len(result_dnlist)
        )

      ContextMenuList.extend([
        form.applAnchor(
          'searchform','Refine Filter',sid,
          [
            ('dn',dn),
            ('searchform_mode','exp'),
            ('search_root',dn),
            ('search_filterstr',search_filterstr),
            ('search_attrs',','.join(read_attrs)),
            ('scope',str(scope)),
          ],
          title='Modify search parameters: filter, search root/scope, etc.',
        ),
        form.applAnchor(
          'search',
          {0:'Raw',1:'Table'}[search_output=='raw'],
          sid,
          [
            ('dn',dn),
            ('search_output',{0:'raw',1:'table'}[search_output=='raw']),
            ('scope',str(scope)),
            ('search_filterstr',search_filterstr),
            ('search_resminindex',str(search_resminindex)),
            ('search_resnumber',str(search_resnumber)),
          ],
          title='Display %s of search results' % (
            {0:'distinguished names',1:'attributes'}[search_output=='raw']
          ),
        ),
      ])

      w2lapp.gui.TopSection(
        sid,outf,form,ls,dn,
        'Search Results',
        w2lapp.gui.MainMenu(sid,form,ls,dn),
        context_menu_list=ContextMenuList
      )

      export_field = w2lapp.form.ExportFormatSelect('search_output')
      export_field.charset = form.accept_charset

      outf.write('<div id="Message" class="Main">\n%s\n%s\n%s\n%s\n' % (
          SearchWarningMsg,
          result_message,
          search_ldap_url.htmlHREF(hrefText='Search by LDAP URL',hrefTarget=None),
          form.formHTML(
            'search','Export',sid,'GET',
            [
              ('dn',dn),
              ('scope',str(scope)),
              ('search_filterstr',search_filterstr),
              ('search_resnumber','0'),
            ],
            extrastr = export_field.inputHTML()+'Incl. op. attrs.:'+w2lapp.form.InclOpAttrsCheckbox('search_opattrs',u'Request operational attributes',default="yes",checked=0).inputHTML(),
            target='web2ldapexport',
          ),
        )
      )

      mailtolist = []
      for r in result_dnlist:
        if is_search_result.has_key(r[0]):
          mailtolist.extend(r[1][1].get('mail',r[1][1].get('rfc822Mailbox',[])))
      if mailtolist:
        mailtolist = [ urllib.quote(m) for m in mailtolist ]
        outf.write('Mail to all <a href="mailto:%s?cc=%s">Cc:-ed</a> - <a href="mailto:?bcc=%s">Bcc:-ed</a>' % (
          mailtolist[0],
          ','.join(mailtolist[1:]),
          ','.join(mailtolist)
        ))

      outf.write('<dl class="SearchResultList">\n')

      for r in result_dnlist:

        if is_search_reference.has_key(r[0]):

          # Display a search continuation (search reference)
          entry = ldap.cidict.cidict({})
          try:
            refUrl = LDAPUrl(r[1][1][0])
          except ValueError,e:
            command_table = []
            result_dd_str='Search reference (NON-LDAP-URI) =&gt; %s' % (w2lapp.gui.utf2display(form.accept_charset,r[1][1][0]))
          else:
            result_dd_str='Search reference =&gt; %s' % (refUrl.htmlHREF(hrefTarget=None))
            if scope==ldap.SCOPE_SUBTREE:
              refUrl.scope = refUrl.scope or scope
              refUrl.filterstr = (unicode(refUrl.filterstr or '',ls.charset) or search_filterstr).encode(form.accept_charset)
              command_table = [
                form.applAnchor(
                  'search','Continue search',
                  {0:sid,1:None}[refUrl.initializeUrl()!=ls.uri],
                  [('ldapurl',refUrl.unparse())],
                  title='Follow this search continuation',
                )
              ]
            else:
              command_table = []
              refUrl.filterstr = search_filterstr
              refUrl.scope=ldap.SCOPE_BASE
              command_table.append(form.applAnchor(
                'read','Read',
                {0:sid,1:None}[refUrl.initializeUrl()!=ls.uri],
                [('ldapurl',refUrl.unparse())],
                title='Display single entry following search continuation',
              ))
              refUrl.scope=ldap.SCOPE_ONELEVEL
              command_table.append(form.applAnchor(
                'search','Down',
                {0:sid,1:None}[refUrl.initializeUrl()!=ls.uri],
                [('ldapurl',refUrl.unparse())],
                title='Descend into tree following search continuation',
              ))

        elif is_search_result.has_key(r[0]):

          # Display a search result with entry's data
          dn,entry = unicode(r[1][0],ls.charset),ldap.cidict.cidict(r[1][1])

          if search_output=='raw':
          
            # Output DN
            result_dd_str=utf2display(form.accept_charset,dn)

          else:

            if entry.has_key('displayName'):

              display_name = unicode(entry['displayName'][0],ls.charset)
              result_dd_str=display_name.encode(form.accept_charset)

            else:

              objectclasses_lower_set = Set([o.lower() for o in entry.get('objectClass',[])])
              tdtemplate_oc = objectclasses_lower_set.intersection(search_tdtemplate_keys_lower)

              if tdtemplate_oc:

                template_attrs = Set([])
                for oc in tdtemplate_oc:
                  template_attrs.update(GrabKeys(search_tdtemplate[oc]).keys)
                tableentry_attrs = template_attrs.intersection(entry.data.keys())
                if tableentry_attrs:
                  # Output entry with the help of pre-defined templates
                  tableentry = msbase.CaseinsensitiveStringKeyDict(default='&nbsp;')
                  for attr in tableentry_attrs:
                    tableentry[attr] = []
                    for value in entry[attr]:
                      tableentry[attr].append(
                        w2lapp.gui.DataStr(sid,form,ls,dn,sub_schema,attr,value,commandbutton=0)
                      )
                    tableentry[attr] = ', '.join(tableentry[attr])
                  tdlist = []
                  for oc in tdtemplate_oc:
                    tdlist.append(search_tdtemplate[oc] % tableentry)
                  result_dd_str='<br>\n'.join(tdlist)
                else:
                  # Output DN
                  result_dd_str=utf2display(form.accept_charset,dn)

              elif search_tablistattrs and entry.has_key(search_tablistattrs[0]):
                tdlist = []
                for attr_type in search_tablistattrs:
                  if entry.has_key(attr_type):
                    tdlist.append(', '.join([
                      w2lapp.gui.DataStr(
                        sid,form,ls,dn,sub_schema,attr_type,value,commandbutton=0
                      )
                      for value in entry[attr_type]
                    ]))
                result_dd_str='<br>\n'.join(filter(None,tdlist))

              else:
                # Output DN
                result_dd_str=utf2display(form.accept_charset,dn)

          # Build the list for link table
          command_table = []

          # A [Read] link is added in any case
          read_title_list = [ dn ]
          for attr_type in (u'description',u'structuralObjectClass'):
            try:
              first_attr_value = unicode(entry[attr_type][0],ls.charset)
            except KeyError:
              pass
            else:
              read_title_list.append(u'%s: %s' % (attr_type,first_attr_value))
          command_table.append(
            form.applAnchor(
              'read','Read',sid,
              [('dn',dn)],
              title=u'\n'.join(read_title_list)
            )
          )

          # Try to determine from entry's attributes if there are subordinates
          hasSubordinates = entry.get('hasSubordinates',['TRUE'])[0].upper()=='TRUE'
          try:
            subordinateCountFlag = int(entry.get('subordinateCount',entry.get('numAllSubordinates',entry.get('msDS-Approx-Immed-Subordinates',['1'])))[0])
          except ValueError:
            subordinateCountFlag = 1

          # If subordinates or unsure a [Down] link is added
          if hasSubordinates and subordinateCountFlag>0:

            down_title_list = ['Browse subordinates']

            # Determine number of direct subordinates
            numSubOrdinates = entry.get('numSubOrdinates',entry.get('subordinateCount',entry.get('countImmSubordinates',entry.get('msDS-Approx-Immed-Subordinates',[None]))))[0]
            if numSubOrdinates!=None:
              numSubOrdinates = int(numSubOrdinates)
              down_title_list.append('direct: %d' % (numSubOrdinates))
            # Determine total number of subordinates
            numAllSubOrdinates = entry.get('numAllSubOrdinates',entry.get('countTotSubordinates',[None]))[0]
            if numAllSubOrdinates!=None:
              numAllSubOrdinates = int(numAllSubOrdinates)
              down_title_list.append('total: %d' % (numAllSubOrdinates))

            command_table.append(form.applAnchor(
                'search','Down',sid,
                [('dn',dn),('scope',str(ldap.SCOPE_ONELEVEL))],
                title='\r\n'.join(down_title_list)
            ))

        else:
          raise ValueError,"LDAP result of invalid type %s." % (repr(r[0]))

        outf.write('<dt class="CommandTable">\n%s\n</dt>\n' % (
            ' | \n'.join(command_table)
          )
        )
        outf.write('<dd class="SearchResultItem">\n%s\n</dd>\n' % (result_dd_str))
      outf.write('</dl>\n</div>\n')

      w2lapp.gui.PrintFooter(outf,form)

    else:
    
      ##############################
      # Empty search results
      ##############################

      ContextMenuList = [
        form.applAnchor(
          'searchform','Refine Filter',sid,
          [
            ('dn',dn),
            ('searchform_mode','exp'),
            ('search_root',dn),
            ('search_filterstr',search_filterstr),
            ('search_attrs',','.join(read_attrs)),
            ('scope',str(scope)),
          ],
        )
      ]

      w2lapp.gui.SimpleMessage(
        sid,outf,form,ls,dn,
        'No Search Results',
        'No Entries found with %s search with filter &quot;%s&quot;.' % (
          ldaputil.base.SEARCH_SCOPE_STR[scope],
          utf2display(form.accept_charset,search_filterstr)
        ),
        main_menu_list=w2lapp.gui.MainMenu(sid,form,ls,dn),
        context_menu_list=ContextMenuList
      )

  elif search_output=='print':

    result_handler.processResults(timeout=ls.timeout)
    result_handler.allResults.sort()

    table=[]
    for r in result_handler.allResults:
      if is_search_result.has_key(r[0]):
        entry = r[1][1]
        objectclasses = entry.get('objectclass',entry.get('objectClass',[]))
        template_oc = list(Set([o.lower() for o in objectclasses]).intersection(
          [s.lower() for s in print_template_str_dict.keys()]
        ))
        if template_oc:
          tableentry = msbase.CaseinsensitiveStringKeyDict(default='&nbsp;')
          attr_list=entry.keys()
          for attr in attr_list:
            tableentry[attr] = ', '.join([
              utf2display(form.accept_charset,attr_value)
              for attr_value in entry[attr]
            ])
          table.append(print_template_str_dict[template_oc[0]] % (tableentry))

    # Output search results as pretty-printable table without buttons
    w2lapp.gui.PrintHeader(sid,outf,form,ls,'Pretty-Printable Search Results')
    outf.write("""<h1>%s</h1>
<table
  border="1"
  rules="rows"
  id="PrintTable"
  summary="Table with search results formatted for printing">
    <colgroup span="%d">
    """ % (utf2display(form.accept_charset,dn),print_cols))
    tdwidth=int(100/print_cols)
    for i in range(print_cols):
      outf.write('<col width="%d%%">\n' % (tdwidth))
    outf.write('</colgroup>\n')
    for i in range(len(table)):
      if i%print_cols==0:
        outf.write('<tr>')
      outf.write('<td width="%d%%">%s</td>' % (tdwidth,table[i]))
      if i%print_cols==print_cols-1:
        outf.write('</tr>\n')
    if len(table)%print_cols!=print_cols-1:
      outf.write('</tr>\n')

    outf.write('</table>\n')

    w2lapp.gui.PrintFooter(outf,form)


  ################################################################
  # Output LDIF or LDIF 1.0
  ################################################################

  elif search_output in ['ldif','ldif1']:

    # Output search results as LDIF
    pyweblib.httphelper.SendHeader(
      outf,'text/plain',charset='UTF-8',
      additional_header={
        'Cache-Control':'private,no-store,no-cache,max-age=0,must-revalidate',
        'Content-Disposition':'inline; filename=web2ldap-export.ldif',
      }
    )
    if search_output=='ldif1':
      result_handler.headerStr = """########################################################################
# LDIF export by web2ldap %s, see http://www.web2ldap.de
# Date and time: %s
# Bind-DN: %s
# LDAP-URL of search:
# %s
########################################################################
version: 1

""" % (
      w2lapp.__version__,
      time.strftime(
        '%A, %Y-%m-%d %H:%M:%S GMT',
        time.gmtime(time.time())
      ),
      repr(ls.who),str(search_ldap_url)
    )
    result_handler.processResults(timeout=ls.timeout)


  ################################################################
  # Output DSML
  ################################################################

  elif search_output=='dsml1':

    # Output search results as DSML in a rather primitive way.
    # (level 1 producer)
    pyweblib.httphelper.SendHeader(
      outf,'text/xml',charset='UTF-8',
      additional_header={
        'Cache-Control':'private',
        'Content-Disposition':'inline; filename=web2ldap-export.dsml'
      }
    )
    result_handler._dsmlWriter._dsml_comment = ('\n'+result_handler._dsmlWriter._indent).join([
        'DSML exported by web2ldap %s, see http://www.web2ldap.de' % (w2lapp.__version__),
        'Date and time: %s' % (time.strftime('%A, %Y-%m-%d %H:%M:%S GMT',time.gmtime(time.time()))),
        'Bind-DN: %s' % (repr(ls.who)),
        'LDAP-URL of search:',
        str(search_ldap_url),
    ])
    result_handler.processResults(timeout=ls.timeout)
